#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Akini Ross (@akinross) <akinross@cisco.com>
# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# Copyright: (c) 2018, Dag Wieers (@dagwieers) <dag@wieers.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_template_contract_filter
short_description: Manage contract filters in schema templates
description:
- Manage contract filters in schema templates on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  contract:
    description:
    - The name of the contract to manage.
    type: str
    required: true
  description:
    description:
    - The description of contract is supported on versions of MSO/NDO that are 3.3 or greater.
    type: str
  contract_display_name:
    description:
    - The name as displayed on the MSO web interface.
    - This defaults to the contract name when unset on creation.
    type: str
  contract_filter_type:
    description:
    - DEPRECATION WARNING, contract_filter_type will not be used anymore and is deduced from filter_type.
    - The type of filters defined in this contract.
    - This defaults to C(both-way) when unset on creation.
    default: both-way
    type: str
    choices: [ both-way, one-way ]
  contract_scope:
    description:
    - The scope of the contract.
    - This defaults to C(vrf) when unset on creation.
    type: str
    choices: [ application-profile, global, tenant, vrf ]
  filter:
    description:
    - The filter to associate with this contract.
    type: str
    aliases: [ name ]
  filter_template:
    description:
    - The template name in which the filter is located.
    type: str
  filter_schema:
    description:
    - The schema name in which the filter is located.
    type: str
  filter_type:
    description:
    - The type of filter to manage.
    - Prior to MSO/NDO 3.3 remove and re-apply contract to change the filter type.
    type: str
    choices: [ both-way, consumer-to-provider, provider-to-consumer ]
    default: both-way
    aliases: [ type ]
  filter_directives:
    description:
    - A list of filter directives.
    type: list
    elements: str
    choices: [ log, none, policy_compression ]
  qos_level:
    description:
    - The Contract QoS Level parameter is supported on versions of MSO/NDO that are 3.3 or greater.
    type: str
    choices: [ unspecified, level1, level2, level3, level4, level5, level6 ]
  action:
    description:
    - The filter action parameter is supported on versions of MSO/NDO that are 3.3 or greater.
    type: str
    choices: [ permit, deny ]
  priority:
    description:
    - The filter priority override parameter is supported on versions of MSO/NDO that are 3.3 or greater.
    type: str
    choices: [ default, lowest_priority, medium_priority, highest_priority ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
seealso:
- module: cisco.mso.mso_schema_template_filter_entry
notes:
- Due to restrictions of the MSO/NDO REST API this module creates contracts when needed, and removes them when the last filter has been removed.
- Due to restrictions of the MSO/NDO REST API concurrent modifications to contract filters can be dangerous and corrupt data.
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new contract filter
  cisco.mso.mso_schema_template_contract_filter:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    contract: Contract 1
    contract_scope: global
    filter: Filter 1
    state: present
  delegate_to: localhost

- name: Remove a contract filter
  cisco.mso.mso_schema_template_contract_filter:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    contract: Contract 1
    filter: Filter 1
    state: absent
  delegate_to: localhost

- name: Query a specific contract filter
  cisco.mso.mso_schema_template_contract_filter:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    contract: Contract 1
    filter: Filter 1
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all contract filters
  cisco.mso.mso_schema_template_contract_filter:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema 1
    template: Template 1
    contract: Contract 1
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.constants import FILTER_KEY_MAP, PRIORITY_MAP


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        template=dict(type="str", required=True),
        contract=dict(type="str", required=True),
        description=dict(type="str"),
        contract_display_name=dict(type="str"),
        contract_scope=dict(type="str", choices=["application-profile", "global", "tenant", "vrf"]),
        # Deprecated input: contract_filter_type is deduced from filter_type
        contract_filter_type=dict(type="str", default="both-way", choices=["both-way", "one-way"]),
        filter=dict(type="str", aliases=["name"]),  # This parameter is not required for querying all objects
        filter_directives=dict(type="list", elements="str", choices=["log", "none", "policy_compression"]),
        filter_template=dict(type="str"),
        filter_schema=dict(type="str"),
        filter_type=dict(type="str", default="both-way", choices=list(FILTER_KEY_MAP), aliases=["type"]),
        qos_level=dict(type="str", choices=["unspecified", "level1", "level2", "level3", "level4", "level5", "level6"]),
        action=dict(type="str", choices=["permit", "deny"]),
        priority=dict(type="str", choices=["default", "lowest_priority", "medium_priority", "highest_priority"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["filter"]],
            ["state", "present", ["filter"]],
        ],
    )

    schema = module.params.get("schema")
    template_name = module.params.get("template").replace(" ", "")
    contract_name = module.params.get("contract")
    contract_display_name = module.params.get("contract_display_name")
    description = module.params.get("description")
    # Deprecated input: contract_filter_type is deduced from filter_type.
    # contract_filter_type = module.params.get('contract_filter_type')
    contract_scope = module.params.get("contract_scope")
    filter_name = module.params.get("filter")
    filter_directives = module.params.get("filter_directives")
    filter_template = module.params.get("filter_template")
    filter_schema = module.params.get("filter_schema")
    filter_type = module.params.get("filter_type")
    filter_action = module.params.get("action")
    filter_priority = module.params.get("priority")
    qos_level = module.params.get("qos_level")

    state = module.params.get("state")

    mso = MSOModule(module)

    # Initialize variables
    ops = []
    filter_obj = None
    filter_key = FILTER_KEY_MAP.get(filter_type)
    filter_template = template_name if filter_template is None else filter_template.replace(" ", "")
    filter_schema = schema if filter_schema is None else filter_schema
    filter_schema_id = mso.lookup_schema(filter_schema)
    contract_filter_type = "bothWay" if filter_type == "both-way" else "oneWay"

    # Set path defaults, when object (contract or filter) is found append /{name} to base paths
    base_contract_path = "/templates/{0}/contracts".format(template_name)
    base_filter_path = "{0}/{1}/{2}".format(base_contract_path, contract_name, filter_key)
    contract_path = "{0}/-".format(base_contract_path)
    filter_path = "{0}/-".format(base_filter_path)

    # Get schema information.
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get template by unique identifier "name".
    template_obj = next((item for item in schema_obj.get("templates") if item.get("name") == template_name), None)
    if not template_obj:
        existing_templates = [t.get("name") for t in schema_obj.get("templates")]
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template_name, ", ".join(existing_templates)))

    filter_ref = mso.filter_ref(schema_id=filter_schema_id, template=filter_template, filter=filter_name)

    # Get contract by unique identifier "name".
    contract_obj = next((item for item in template_obj.get("contracts") if item.get("name") == contract_name), None)
    if contract_obj:
        if contract_obj.get("filterType") != contract_filter_type:
            mso.fail_json(
                msg="Current filter type '{0}' for contract '{1}' is not allowed to change to '{2}'.".format(
                    contract_obj.get("filterType"), contract_name, contract_filter_type
                )
            )
        contract_path = "{0}/{1}".format(base_contract_path, contract_name)
        if filter_name:
            # Get filter by unique identifier "filterRef".
            filter_obj = next((item for item in contract_obj.get(filter_key) if item.get("filterRef") == filter_ref), None)
            if filter_obj:
                filter_path = "{0}/{1}".format(base_filter_path, filter_name)
                mso.update_filter_obj(contract_obj, filter_obj, filter_type)
                mso.existing = filter_obj

    if state == "query":
        if not contract_obj:
            existing_contracts = [c.get("name") for c in template_obj.get("contracts")]
            mso.fail_json(msg="Provided contract '{0}' does not exist. Existing contracts: {1}".format(contract_name, ", ".join(existing_contracts)))

        # If filter name is not provided, provide overview of all filter objects for the filter type.
        if not filter_name:
            mso.existing = contract_obj.get(filter_key)
            for filter_obj in mso.existing:
                mso.update_filter_obj(contract_obj, filter_obj, filter_type)

        elif not mso.existing:
            mso.fail_json(msg="FilterRef '{filter_ref}' not found".format(filter_ref=filter_ref))

        mso.exit_json()

    mso.previous = mso.existing

    if state == "absent":
        # Contracts need at least one filter left, remove contract if remove would lead 0 filters remaining.
        if contract_obj:
            if len(contract_obj.get(filter_key)) == 1:
                mso.existing = {}
                ops.append(dict(op="remove", path=contract_path))
            elif len(contract_obj.get(filter_key)) > 1:
                mso.existing = {}
                ops.append(dict(op="remove", path=filter_path))

    elif state == "present":
        contract_scope = "context" if contract_scope == "vrf" else contract_scope

        # Initialize "present" state filter variables
        if not filter_directives:
            # Avoid validation error: "Bad Request: (0)(1)(0) 'directives' is undefined on object
            if not filter_obj:
                filter_directives = ["none"]
            else:
                filter_directives = filter_obj.get("directives", ["none"])

        elif "policy_compression" in filter_directives:
            filter_directives[filter_directives.index("policy_compression")] = "no_stats"
        filter_payload = dict(
            filterRef=dict(
                filterName=filter_name,
                templateName=filter_template,
                schemaId=filter_schema_id,
            ),
            directives=filter_directives,
        )
        if filter_action:
            filter_payload.update(action=filter_action)
        if filter_action == "deny" and filter_priority:
            filter_payload.update(priorityOverride=PRIORITY_MAP.get(filter_priority))

        # If contract exist the operation should be set to replace else operation is add to create new contract.
        if contract_obj:
            if contract_display_name:
                ops.append(dict(op="replace", path=contract_path + "/displayName", value=contract_display_name))
            # Conditional statement 'description == ""' is needed to allow setting the description back to empty string.
            if description or description == "":
                ops.append(dict(op="replace", path=contract_path + "/description", value=description))
            if qos_level:
                # Conditional statement is needed to determine if "prio" exist in contract object.
                # An object can be created in 3.3 higher version without prio via the API.
                # In the GUI a default is set to "unspecified" and thus prio is always configured via GUI.
                # We can't set a default of "unspecified" because prior to version 3.3 qos_level is not supported,
                #  thus the logic is needed for both add and replace operation
                if contract_obj.get("prio"):
                    ops.append(dict(op="replace", path=contract_path + "/prio", value=qos_level))
                else:
                    ops.append(dict(op="add", path=contract_path + "/prio", value=qos_level))
            if contract_scope:
                ops.append(dict(op="replace", path=contract_path + "/scope", value=contract_scope))

            # If filter exist the operation should be set to replace else operation is add to create new filter.
            if filter_obj:
                ops.append(dict(op="replace", path=filter_path, value=filter_payload))
            else:
                ops.append(dict(op="add", path=filter_path, value=filter_payload))

        else:
            contract_display_name = contract_display_name if contract_display_name else contract_name
            # If contract_scope is not provided default to context to match GUI behaviour on create new contract.
            contract_scope = "context" if contract_scope is None else contract_scope
            contract_payload = dict(name=contract_name, displayName=contract_display_name, filterType=contract_filter_type, scope=contract_scope)
            if description:
                contract_payload.update(description=description)
            if qos_level:
                contract_payload.update(prio=qos_level)
            if filter_key == "filterRelationships":
                contract_payload.update(filterRelationships=[filter_payload])
            elif filter_key == "filterRelationshipsConsumerToProvider":
                contract_payload.update(filterRelationshipsConsumerToProvider=[filter_payload])
            elif filter_key == "filterRelationshipsProviderToConsumer":
                contract_payload.update(filterRelationshipsProviderToConsumer=[filter_payload])
            ops.append(dict(op="add", path=contract_path, value=contract_payload))

        mso.sanitize(filter_payload, collate=True, unwanted=["filterType", "contractScope", "contractFilterType"])

        # Update existing with filter (mso.sent) and contract information.
        mso.existing = mso.sent
        mso.existing["displayName"] = contract_display_name if contract_display_name else contract_obj.get("displayName")
        mso.existing["filterType"] = filter_type
        mso.existing["contractScope"] = contract_scope if contract_scope else contract_obj.get("scope")
        mso.existing["contractFilterType"] = contract_filter_type
        # Conditional statement 'description == ""' is needed to allow setting the description back to empty string.
        if description or (contract_obj and (contract_obj.get("description") or contract_obj.get("description") == "")):
            mso.existing["description"] = description if description or description == "" else contract_obj.get("description")
        # Conditional statement to check qos_level is defined or is present in the contract object.
        # qos_level is not supported prior to 3.3 thus this check in place, GUI uses default of "unspecified" from 3.3.
        # When default of "unspecified" is set, conditional statement can be simplified since "prio" always present.
        if qos_level or (contract_obj and contract_obj.get("prio")):
            mso.existing["prio"] = qos_level if qos_level else contract_obj.get("prio")

    if not module.check_mode and mso.existing != mso.previous:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = """
---
module: digital_ocean_domain_record
author: "Adam Papai (@woohgit)"
version_added: 1.1.0
short_description: Manage DigitalOcean domain records
description:
    - Create/delete a domain record in DigitalOcean.
options:
  state:
    description:
     - Indicate desired state of the target.
    default: present
    choices: [ present, absent ]
    type: str
  record_id:
    description:
      - Used with C(force_update=yes) and C(state='absent') to update or delete a specific record.
    type: int
  force_update:
    description:
        - If there is already a record with the same C(name) and C(type) force update it.
    default: false
    type: bool
  domain:
    description:
     - Name of the domain.
    required: true
    type: str
  type:
    description:
     - The type of record you would like to create.
    choices: [ A, AAAA, CNAME, MX, TXT, SRV, NS, CAA ]
    type: str
  data:
    description:
     - This is the value of the record, depending on the record type.
    type: str
  name:
    description:
     - Required for C(A, AAAA, CNAME, TXT, SRV) records. The host name, alias, or service being defined by the record.
    default: "@"
    type: str
  priority:
    description:
     - The priority of the host for C(SRV, MX) records).
    type: int
  port:
    description:
     - The port that the service is accessible on for SRV records only.
    type: int
  weight:
    description:
     - The weight of records with the same priority for SRV records only.
    type: int
  ttl:
    description:
     - Time to live for the record, in seconds.
    default: 1800
    type: int
  flags:
    description:
     - An unsignedinteger between 0-255 used for CAA records.
    type: int
  tag:
    description:
     - The parameter tag for CAA records.
    choices: [ issue, wildissue, iodef ]
    type: str
  oauth_token:
    description:
     - DigitalOcean OAuth token. Can be specified in C(DO_API_KEY), C(DO_API_TOKEN), or C(DO_OAUTH_TOKEN) environment variables
    aliases: ['API_TOKEN']
    type: str

notes:
  - Version 2 of DigitalOcean API is used.
  - The number of requests that can be made through the API is currently limited to 5,000 per hour per OAuth token.
"""

EXAMPLES = """
- name: Create default A record for example.com
  community.digitalocean.digital_ocean_domain_record:
    state: present
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    type: A
    name: "@"
    data: 127.0.0.1

- name: Create A record for www
  community.digitalocean.digital_ocean_domain_record:
    state: present
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    type: A
    name: www
    data: 127.0.0.1

- name: Update A record for www based on name/type/data
  community.digitalocean.digital_ocean_domain_record:
    state: present
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    type: A
    name: www
    data: 127.0.0.2
    force_update: true

- name: Update A record for www based on record_id
  community.digitalocean.digital_ocean_domain_record:
    state: present
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    record_id: 123456
    type: A
    name: www
    data: 127.0.0.2
    force_update: true

- name: Remove www record based on name/type/data
  community.digitalocean.digital_ocean_domain_record:
    state: absent
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    type: A
    name: www
    data: 127.0.0.1

- name: Remove www record based on record_id
  community.digitalocean.digital_ocean_domain_record:
    state: absent
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    record_id: 1234567

- name: Create CNAME records for www, git and status subdomains
  community.digitalocean.digital_ocean_domain_record:
    state: present
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    type: CNAME
    name: "{{ item }}"
    data: example.com
  with_items:
    - www
    - git
    - status

- name: Create MX record with priority 10 for example.com
  community.digitalocean.digital_ocean_domain_record:
    state: present
    oauth_token: "{{ lookup('ansible.builtin.env', 'DO_API_TOKEN') }}"
    domain: example.com
    type: MX
    data: mail1.example.com
    priority: 10
"""

RETURN = r"""
data:
    description: a DigitalOcean Domain Record
    returned: success
    type: dict
    sample: {
        "id": 3352896,
        "type": "CNAME",
        "name": "www",
        "data": "192.168.0.1",
        "priority": 10,
        "port": 5556,
        "ttl": 3600,
        "weight": 10,
        "flags": 16,
        "tag": "issue"
    }
"""


from ansible.module_utils.basic import env_fallback
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.community.digitalocean.plugins.module_utils.digital_ocean import (
    DigitalOceanHelper,
)


class DigitalOceanDomainRecordManager(DigitalOceanHelper, object):
    def __init__(self, module):
        super(DigitalOceanDomainRecordManager, self).__init__(module)
        self.module = module
        self.domain = module.params.get("domain").lower()
        self.records = self.__get_all_records()
        self.payload = self.__build_payload()
        self.force_update = module.params.get("force_update", False)
        self.record_id = module.params.get("record_id", None)

    def check_credentials(self):
        # Check if oauth_token is valid or not
        response = self.get("account")
        if response.status_code == 401:
            self.module.fail_json(
                msg="Failed to login using oauth_token, please verify validity of oauth_token"
            )

    def verify_domain(self):
        # URL https://api.digitalocean.com/v2/domains/[NAME]
        response = self.get("domains/%s" % self.domain)
        status_code = response.status_code
        json = response.json

        if status_code not in (200, 404):
            self.module.fail_json(
                msg="Error getting domain [%(status_code)s: %(json)s]"
                % {"status_code": status_code, "json": json}
            )
        elif status_code == 404:
            self.module.fail_json(
                msg="No domain named '%s' found. Please create a domain first"
                % self.domain
            )

    def __get_all_records(self):
        records = []
        page = 1
        while True:
            # GET /v2/domains/$DOMAIN_NAME/records
            response = self.get(
                "domains/%(domain)s/records?page=%(page)s"
                % {"domain": self.domain, "page": page}
            )
            status_code = response.status_code
            json = response.json

            if status_code != 200:
                self.module.fail_json(
                    msg="Error getting domain records [%(status_code)s: %(json)s]"
                    % {"status_code": status_code, "json": json}
                )

            for record in json["domain_records"]:
                records.append(dict([(str(k), v) for k, v in record.items()]))

            if "pages" in json["links"] and "next" in json["links"]["pages"]:
                page += 1
            else:
                break

        return records

    def __normalize_data(self):
        # for the MX, CNAME, SRV, CAA records make sure the data ends with a dot
        if (
            self.payload["type"] in ["CNAME", "MX", "SRV", "CAA"]
            and self.payload["data"] != "@"
            and not self.payload["data"].endswith(".")
        ):
            data = "%s." % self.payload["data"]
        else:
            data = self.payload["data"]

        return data

    def __find_record_by_id(self, record_id):
        for record in self.records:
            if record["id"] == record_id:
                return record
        return None

    def __get_matching_records(self):
        """Collect exact and similar records

        It returns an exact record if there is any match along with the record_id.
        It also returns multiple records if there is no exact match
        """

        # look for exactly the same record used by (create, delete)
        for record in self.records:
            r = dict(record)
            del r["id"]
            # python3 does not have cmp so let's use the official workaround
            if r == self.payload:
                return r, record["id"], None

        # look for similar records used by (update)
        similar_records = []
        for record in self.records:
            if (
                record["type"] == self.payload["type"]
                and record["name"] == self.payload["name"]
            ):
                similar_records.append(record)

        if similar_records:
            return None, None, similar_records

        # if no exact neither similar records
        return None, None, None

    def __create_record(self):
        # before data comparison, we need to make sure that
        # the payload['data'] is not normalized, but
        # during create/update digitalocean expects normalized data
        self.payload["data"] = self.__normalize_data()

        # POST /v2/domains/$DOMAIN_NAME/records
        response = self.post("domains/%s/records" % self.domain, data=self.payload)
        status_code = response.status_code
        json = response.json
        if status_code == 201:
            changed = True
            return changed, json["domain_record"]
        else:
            self.module.fail_json(
                msg="Error creating domain record [%(status_code)s: %(json)s]"
                % {"status_code": status_code, "json": json}
            )

    def create_or_update_record(self):
        # if record_id is given we need to update the record no matter what
        if self.record_id:
            changed, result = self.__update_record(self.record_id)
            return changed, result

        record, record_id, similar_records = self.__get_matching_records()

        # create the record if no similar or exact record were found
        if not record and not similar_records:
            changed, result = self.__create_record()
            return changed, result

        # no exact match, but we have similar records
        # so if force_update == True we should update it
        if not record and similar_records:
            # if we have 1 similar record
            if len(similar_records) == 1:
                # update if we were told to do it so
                if self.force_update:
                    record_id = similar_records[0]["id"]
                    changed, result = self.__update_record(record_id)
                # if no update was given, create it
                else:
                    changed, result = self.__create_record()
                return changed, result
            # we have multiple similar records, bun not exact match
            else:
                # we have multiple similar records, can't decide what to do
                if self.force_update:
                    self.module.fail_json(
                        msg="Can't update record, too many similar records: %s"
                        % similar_records
                    )
                # create it
                else:
                    changed, result = self.__create_record()
                return changed, result
        # record matches
        else:
            changed = False
            result = "Record has been already created"
            return changed, result

    def __update_record(self, record_id):
        # before data comparison, we need to make sure that
        # the payload['data'] is not normalized, but
        # during create/update digitalocean expects normalized data
        self.payload["data"] = self.__normalize_data()

        # double check if the record exist
        record = self.__find_record_by_id(record_id)

        # record found
        if record:
            # PUT /v2/domains/$DOMAIN_NAME/records/$RECORD_ID
            response = self.put(
                "domains/%(domain)s/records/%(record_id)s"
                % {"domain": self.domain, "record_id": record_id},
                data=self.payload,
            )
            status_code = response.status_code
            json = response.json
            if status_code == 200:
                changed = True
                return changed, json["domain_record"]
            else:
                self.module.fail_json(
                    msg="Error updating domain record [%(status_code)s: %(json)s]"
                    % {"status_code": status_code, "json": json}
                )
        # recond not found
        else:
            self.module.fail_json(
                msg="Error updating domain record. Record does not exist. [%s]"
                % record_id
            )

    def __build_payload(self):
        payload = dict(
            data=self.module.params.get("data"),
            flags=self.module.params.get("flags"),
            name=self.module.params.get("name"),
            port=self.module.params.get("port"),
            priority=self.module.params.get("priority"),
            type=self.module.params.get("type"),
            tag=self.module.params.get("tag"),
            ttl=self.module.params.get("ttl"),
            weight=self.module.params.get("weight"),
        )

        # DigitalOcean stores every data in lowercase except TXT
        if payload["type"] != "TXT" and payload["data"]:
            payload["data"] = payload["data"].lower()

        # digitalocean stores data: '@' if the data=domain
        if payload["data"] == self.domain:
            payload["data"] = "@"

        return payload

    def delete_record(self):
        # if record_id is given, try to find the record based on the id
        if self.record_id:
            record = self.__find_record_by_id(self.record_id)
            record_id = self.record_id
        # if no record_id is given, try to a single matching record
        else:
            record, record_id, similar_records = self.__get_matching_records()
            if not record and similar_records:
                if len(similar_records) == 1:
                    record, record_id = similar_records[0], similar_records[0]["id"]
                else:
                    self.module.fail_json(
                        msg="Can't delete record, too many similar records: %s"
                        % similar_records
                    )
        # record was not found, we're done
        if not record:
            changed = False
            return changed, record
        # record found, lets delete it
        else:
            # DELETE /v2/domains/$DOMAIN_NAME/records/$RECORD_ID.
            response = self.delete(
                "domains/%(domain)s/records/%(id)s"
                % {"domain": self.domain, "id": record_id}
            )
            status_code = response.status_code
            json = response.json
            if status_code == 204:
                changed = True
                msg = "Successfully deleted %s" % record["name"]
                return changed, msg
            else:
                self.module.fail_json(
                    msg="Error deleting domain record. [%(status_code)s: %(json)s]"
                    % {"status_code": status_code, "json": json}
                )


def main():
    module = AnsibleModule(
        argument_spec=dict(
            state=dict(choices=["present", "absent"], default="present"),
            oauth_token=dict(
                aliases=["API_TOKEN"],
                no_log=True,
                fallback=(
                    env_fallback,
                    ["DO_API_TOKEN", "DO_API_KEY", "DO_OAUTH_TOKEN"],
                ),
            ),
            force_update=dict(type="bool", default=False),
            record_id=dict(type="int"),
            domain=dict(type="str", required=True),
            type=dict(choices=["A", "AAAA", "CNAME", "MX", "TXT", "SRV", "NS", "CAA"]),
            name=dict(type="str", default="@"),
            data=dict(type="str"),
            priority=dict(type="int"),
            port=dict(type="int"),
            weight=dict(type="int"),
            ttl=dict(type="int", default=1800),
            tag=dict(choices=["issue", "wildissue", "iodef"]),
            flags=dict(type="int"),
        ),
        # TODO
        # somehow define the absent requirements: record_id OR ('name', 'type', 'data')
        required_if=[("state", "present", ("type", "name", "data"))],
    )

    manager = DigitalOceanDomainRecordManager(module)

    # verify credentials and domain
    manager.check_credentials()
    manager.verify_domain()

    state = module.params.get("state")

    if state == "present":
        changed, result = manager.create_or_update_record()
    elif state == "absent":
        changed, result = manager.delete_record()

    module.exit_json(changed=changed, result=result)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = '''
    author: Unknown (!UNKNOWN)
    name: nmap
    short_description: Uses nmap to find hosts to target
    description:
        - Uses a YAML configuration file with a valid YAML extension.
    extends_documentation_fragment:
      - constructed
      - inventory_cache
    requirements:
      - nmap CLI installed
    options:
        plugin:
            description: token that ensures this is a source file for the 'nmap' plugin.
            required: true
            choices: ['nmap', 'community.general.nmap']
        sudo:
            description: Set to V(true) to execute a C(sudo nmap) plugin scan.
            version_added: 4.8.0
            default: false
            type: boolean
        address:
            description: Network IP or range of IPs to scan, you can use a simple range (10.2.2.15-25) or CIDR notation.
            required: true
            env:
                - name: ANSIBLE_NMAP_ADDRESS
                  version_added: 6.6.0
        exclude:
            description:
              - List of addresses to exclude.
              - For example V(10.2.2.15-25) or V(10.2.2.15,10.2.2.16).
            type: list
            elements: string
            env:
                - name: ANSIBLE_NMAP_EXCLUDE
                  version_added: 6.6.0
        port:
            description:
                - Only scan specific port or port range (C(-p)).
                - For example, you could pass V(22) for a single port, V(1-65535) for a range of ports,
                  or V(U:53,137,T:21-25,139,8080,S:9) to check port 53 with UDP, ports 21-25 with TCP, port 9 with SCTP, and ports 137, 139, and 8080 with all.
            type: string
            version_added: 6.5.0
        ports:
            description: Enable/disable scanning ports.
            type: boolean
            default: true
        ipv4:
            description: use IPv4 type addresses
            type: boolean
            default: true
        ipv6:
            description: use IPv6 type addresses
            type: boolean
            default: true
        udp_scan:
            description:
                - Scan via UDP.
                - Depending on your system you might need O(sudo=true) for this to work.
            type: boolean
            default: false
            version_added: 6.1.0
        icmp_timestamp:
            description:
                - Scan via ICMP Timestamp (C(-PP)).
                - Depending on your system you might need O(sudo=true) for this to work.
            type: boolean
            default: false
            version_added: 6.1.0
        open:
            description: Only scan for open (or possibly open) ports.
            type: boolean
            default: false
            version_added: 6.5.0
        dns_resolve:
            description: Whether to always (V(true)) or never (V(false)) do DNS resolution.
            type: boolean
            default: false
            version_added: 6.1.0
        use_arp_ping:
            description: Whether to always (V(true)) use the quick ARP ping or (V(false)) a slower but more reliable method.
            type: boolean
            default: true
            version_added: 7.4.0
    notes:
        - At least one of ipv4 or ipv6 is required to be True, both can be True, but they cannot both be False.
        - 'TODO: add OS fingerprinting'
'''
EXAMPLES = '''
# inventory.config file in YAML format
plugin: community.general.nmap
strict: false
address: 192.168.0.0/24


# a sudo nmap scan to fully use nmap scan power.
plugin: community.general.nmap
sudo: true
strict: false
address: 192.168.0.0/24

# an nmap scan specifying ports and classifying results to an inventory group
plugin: community.general.nmap
address: 192.168.0.0/24
exclude: 192.168.0.1, web.example.com
port: 22, 443
groups:
  web_servers: "ports | selectattr('port', 'equalto', '443')"
'''

import os
import re

from subprocess import Popen, PIPE

from ansible import constants as C
from ansible.errors import AnsibleParserError
from ansible.module_utils.common.text.converters import to_native, to_text
from ansible.plugins.inventory import BaseInventoryPlugin, Constructable, Cacheable
from ansible.module_utils.common.process import get_bin_path
from ansible.utils.unsafe_proxy import wrap_var as make_unsafe


class InventoryModule(BaseInventoryPlugin, Constructable, Cacheable):

    NAME = 'community.general.nmap'
    find_host = re.compile(r'^Nmap scan report for ([\w,.,-]+)(?: \(([\w,.,:,\[,\]]+)\))?')
    find_port = re.compile(r'^(\d+)/(\w+)\s+(\w+)\s+(\w+)')

    def __init__(self):
        self._nmap = None
        super(InventoryModule, self).__init__()

    def _populate(self, hosts):
        # Use constructed if applicable
        strict = self.get_option('strict')

        for host in hosts:
            host = make_unsafe(host)
            hostname = host['name']
            self.inventory.add_host(hostname)
            for var, value in host.items():
                self.inventory.set_variable(hostname, var, value)

            # Composed variables
            self._set_composite_vars(self.get_option('compose'), host, hostname, strict=strict)

            # Complex groups based on jinja2 conditionals, hosts that meet the conditional are added to group
            self._add_host_to_composed_groups(self.get_option('groups'), host, hostname, strict=strict)

            # Create groups based on variable values and add the corresponding hosts to it
            self._add_host_to_keyed_groups(self.get_option('keyed_groups'), host, hostname, strict=strict)

    def verify_file(self, path):

        valid = False
        if super(InventoryModule, self).verify_file(path):
            file_name, ext = os.path.splitext(path)

            if not ext or ext in C.YAML_FILENAME_EXTENSIONS:
                valid = True

        return valid

    def parse(self, inventory, loader, path, cache=True):

        try:
            self._nmap = get_bin_path('nmap')
        except ValueError as e:
            raise AnsibleParserError('nmap inventory plugin requires the nmap cli tool to work: {0}'.format(to_native(e)))

        super(InventoryModule, self).parse(inventory, loader, path, cache=cache)

        self._read_config_data(path)

        cache_key = self.get_cache_key(path)

        # cache may be True or False at this point to indicate if the inventory is being refreshed
        # get the user's cache option too to see if we should save the cache if it is changing
        user_cache_setting = self.get_option('cache')

        # read if the user has caching enabled and the cache isn't being refreshed
        attempt_to_read_cache = user_cache_setting and cache
        # update if the user has caching enabled and the cache is being refreshed; update this value to True if the cache has expired below
        cache_needs_update = user_cache_setting and not cache

        if attempt_to_read_cache:
            try:
                results = self._cache[cache_key]
            except KeyError:
                # This occurs if the cache_key is not in the cache or if the cache_key expired, so the cache needs to be updated
                cache_needs_update = True

        if not user_cache_setting or cache_needs_update:
            # setup command
            cmd = [self._nmap]

            if self.get_option('sudo'):
                cmd.insert(0, 'sudo')

            if self.get_option('port'):
                cmd.append('-p')
                cmd.append(self.get_option('port'))

            if not self.get_option('ports'):
                cmd.append('-sP')

            if self.get_option('ipv4') and not self.get_option('ipv6'):
                cmd.append('-4')
            elif self.get_option('ipv6') and not self.get_option('ipv4'):
                cmd.append('-6')
            elif not self.get_option('ipv6') and not self.get_option('ipv4'):
                raise AnsibleParserError('One of ipv4 or ipv6 must be enabled for this plugin')

            if self.get_option('exclude'):
                cmd.append('--exclude')
                cmd.append(','.join(self.get_option('exclude')))

            if self.get_option('dns_resolve'):
                cmd.append('-n')

            if self.get_option('udp_scan'):
                cmd.append('-sU')

            if self.get_option('icmp_timestamp'):
                cmd.append('-PP')

            if self.get_option('open'):
                cmd.append('--open')

            if not self.get_option('use_arp_ping'):
                cmd.append('--disable-arp-ping')

            cmd.append(self.get_option('address'))
            try:
                # execute
                p = Popen(cmd, stdout=PIPE, stderr=PIPE)
                stdout, stderr = p.communicate()
                if p.returncode != 0:
                    raise AnsibleParserError('Failed to run nmap, rc=%s: %s' % (p.returncode, to_native(stderr)))

                # parse results
                host = None
                ip = None
                ports = []
                results = []

                try:
                    t_stdout = to_text(stdout, errors='surrogate_or_strict')
                except UnicodeError as e:
                    raise AnsibleParserError('Invalid (non unicode) input returned: %s' % to_native(e))

                for line in t_stdout.splitlines():
                    hits = self.find_host.match(line)
                    if hits:
                        if host is not None and ports:
                            results[-1]['ports'] = ports

                        # if dns only shows arpa, just use ip instead as hostname
                        if hits.group(1).endswith('.in-addr.arpa'):
                            host = hits.group(2)
                        else:
                            host = hits.group(1)

                        # if no reverse dns exists, just use ip instead as hostname
                        if hits.group(2) is not None:
                            ip = hits.group(2)
                        else:
                            ip = hits.group(1)

                        if host is not None:
                            # update inventory
                            results.append(dict())
                            results[-1]['name'] = host
                            results[-1]['ip'] = ip
                            ports = []
                        continue

                    host_ports = self.find_port.match(line)
                    if host is not None and host_ports:
                        ports.append({'port': host_ports.group(1),
                                      'protocol': host_ports.group(2),
                                      'state': host_ports.group(3),
                                      'service': host_ports.group(4)})
                        continue

                # if any leftovers
                if host and ports:
                    results[-1]['ports'] = ports

            except Exception as e:
                raise AnsibleParserError("failed to parse %s: %s " % (to_native(path), to_native(e)))

        if cache_needs_update:
            self._cache[cache_key] = results

        self._populate(results)



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

"""
The module file for nxos_logging_global
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: nxos_logging_global
short_description: Logging resource module.
description:
- This module manages logging configuration on devices running Cisco NX-OS.
version_added: 2.5.0
notes:
- Tested against NX-OS 9.3.6 on Cisco Nexus Switches.
- Limited Support for Cisco MDS
- This module works with connection C(network_cli) and C(httpapi).
- Tested against Cisco MDS NX-OS 9.2(2) with connection C(network_cli).
author: Nilashish Chakraborty (@NilashishC)
options:
  running_config:
    description:
    - This option is used only with state I(parsed).
    - The value of this option should be the output received from the NX-OS device
      by executing the command B(show running-config | include logging).
    - The state I(parsed) reads the configuration from C(running_config) option and
      transforms it into Ansible structured data as per the resource module's argspec
      and the value is then returned in the I(parsed) key within the result.
    type: str
  config:
    description: A dictionary of logging configuration.
    type: dict
    suboptions:
      console:
        description: Set console logging parameters.
        type: dict
        suboptions:
          state:
            description: Enable or disable monitor logging.
            type: str
            choices: ["enabled", "disabled"]
          severity: &sev
            description:  Set severity severity for console.
            type: str
            choices:
            - emergency
            - alert
            - critical
            - error
            - warning
            - notification
            - informational
            - debugging
      event:
        description: Interface events.
        type: dict
        suboptions:
          link_status:
            description: UPDOWN and CHANGE messages.
            type: dict
            suboptions: &event
              enable:
                description: To enable logging overriding port severity configuration.
                type: bool
              default:
                description: Default logging configuration used by interfaces not explicitly configured.
                type: bool
          trunk_status:
            description: TRUNK status messages.
            type: dict
            suboptions: *event
      history:
        description: Modifies severity severity or size for history table.
        type: dict
        suboptions:
          severity: *sev
          size:
            description: Set history table size.
            type: int
      ip:
        description:
          - IP configuration.
          - This option is unsupported on MDS switches.
        type: dict
        suboptions:
          access_list:
            description: Access-List.
            type: dict
            suboptions:
              cache:
                description: Set caching settings.
                type: dict
                suboptions:
                  entries:
                    description: Maximum number of log entries cached in software.
                    type: int
                  interval:
                    description: Log-update interval (in sec).
                    type: int
                  threshold:
                    description: Log-update threshold (number of hits)
                    type: int
              detailed:
                description: Detailed ACL information.
                type: bool
              include:
                description: Include additional fields in syslogs.
                type: dict
                suboptions:
                  sgt:
                    description: Include source group tag info in syslogs.
                    type: bool
      facilities:
        description: Facility parameter for syslog messages.
        type: list
        elements: dict
        suboptions:
          facility:
            description: Facility name.
            type: str
          severity: *sev
      logfile:
        description: Set file logging.
        type: dict
        suboptions:
          state:
            description: Enable or disable logfile.
            type: str
            choices: ["enabled", "disabled"]
          name:
            description: Logfile name.
            type: str
          severity: *sev
          persistent_threshold:
            description:
              - Set persistent logging utilization alert threshold in percentage.
              - This option is unsupported on MDS switches.
            type: int
          size:
            description: Enter the logfile size in bytes.
            type: int
      module:
        description: Set module(linecard) logging.
        type: dict
        suboptions:
          state:
            description: Enable or disable module logging.
            type: str
            choices: ["enabled", "disabled"]
          severity: *sev
      monitor:
        description: Set terminal line(monitor) logging severity.
        type: dict
        suboptions:
          state:
            description: Enable or disable monitor logging.
            type: str
            choices: ["enabled", "disabled"]
          severity: *sev
      origin_id:
        description: Enable origin information for Remote Syslog Server.
        type: dict
        suboptions:
          hostname:
            description:
              - Use hostname as origin-id of logging messages.
              - This option is mutually exclusive with I(ip) and I(string).
            type: bool
          ip:
            description:
              - Use ip address as origin-id of logging messages.
              - This option is mutually exclusive with I(hostname) and I(string).
            type: str
          string:
            description:
              - Use text string as origin-id of logging messages.
              - This option is mutually exclusive with I(hostname) and I(ip).
            type: str
      rate_limit:
        description: Enable or disable rate limit for log messages.
        type: str
        choices: ["enabled", "disabled"]
      rfc_strict:
        description:
          - Set RFC to which messages should compliant.
          - Syslogs will be compliant to RFC 5424.
          - This option is unsupported on MDS switches.
        type: bool
      hosts:
        description: Enable forwarding to Remote Syslog Servers.
        type: list
        elements: dict
        suboptions:
          host:
            description: Hostname/IPv4/IPv6 address of the Remote Syslog Server.
            type: str
          severity: *sev
          facility:
            description: Facility to use when forwarding to server.
            type: str
          port:
            description: Destination Port when forwarding to remote server.
            type: int
          secure:
            description: Enable secure connection to remote server.
            type: dict
            suboptions:
              trustpoint:
                description: Trustpoint configuration.
                type: dict
                suboptions:
                  client_identity:
                    description:
                      - Client Identity certificate for mutual authentication.
                      - Trustpoint to use for client certificate authentication.
                    type: str
          use_vrf:
            description:
              - Display per-VRF information.
              - This option is unsupported on MDS switches.
            type: str
      source_interface:
        description:
          - Enable Source-Interface for Remote Syslog Server.
          - This option is unsupported on MDS switches.
        type: str
      timestamp:
        description: Set logging timestamp granularity.
        type: str
        choices: ["microseconds", "milliseconds", "seconds"]
  state:
    description:
    - The state the configuration should be left in.
    - The states I(replaced) and I(overridden) have identical
      behaviour for this module.
    - Refer to examples for more details.
    type: str
    choices:
    - merged
    - replaced
    - overridden
    - deleted
    - parsed
    - gathered
    - rendered
    default: merged
"""
EXAMPLES = """
# Using merged

# Before state:
# -------------
# nxos-9k-rdo# show running-config | include logging
# nxos-9k-rdo#

- name: Merge the provided configuration with the existing running configuration
  cisco.nxos.nxos_logging_global:
    config:
      console:
        severity: error
      monitor:
        severity: warning
      ip:
        access_list:
          cache:
            entries: 16384
            interval: 200
            threshold: 5000
      facilities:
        - facility: auth
          severity: critical
        - facility: ospfv3
          severity: alert
        - facility: ftp
          severity: informational
      hosts:
        - host: 203.0.113.100
          severity: alert
          use_vrf: management
        - host: 203.0.113.101
          severity: error
          facility: local6
          use_vrf: default
      origin_id:
        hostname: True

# Task output
# -------------
#  before: {}
#
#  commands:
#    - "logging console 3"
#    - "logging monitor 4"
#    - "logging ip access-list cache entries 16384"
#    - "logging ip access-list cache interval 200"
#    - "logging ip access-list cache threshold 5000"
#    - "logging severity auth 2"
#    - "logging severity ospfv3 1"
#    - "logging severity ftp 6"
#    - "logging server 203.0.113.100 1 use-vrf management"
#    - "logging server 203.0.113.101 3 facility local6 use-vrf default"
#    - "logging origin-id hostname"
#
# after:
#   console:
#      severity: error
#    facilities:
#      - facility: auth
#        severity: critical
#      - facility: ftp
#        severity: informational
#      - facility: ospfv3
#        severity: alert
#    ip:
#      access_list:
#        cache:
#          entries: 16384
#          interval: 200
#          threshold: 5000
#    monitor:
#      severity: warning
#    origin_id:
#      hostname: true
#    hosts:
#      - severity: alert
#        host: 203.0.113.100
#        use_vrf: management
#      - facility: local6
#        severity: error
#        host: 203.0.113.101
#        use_vrf: default

# After state:
# ------------
# nxos-9k-rdo# show running-config | include logging
# logging console 3
# logging monitor 4
# logging ip access-list cache entries 16384
# logging ip access-list cache interval 200
# logging ip access-list cache threshold 5000
# logging severity auth 2
# logging severity ospfv3 1
# logging severity ftp 6
# logging origin-id hostname
# logging server 203.0.113.100 1 use-vrf management
# logging server 203.0.113.101 3 use-vrf default facility local6

# Using replaced

# Before state:
# ------------
# nxos-9k-rdo# show running-config | include logging
# logging console 3
# logging monitor 4
# logging ip access-list cache entries 16384
# logging ip access-list cache interval 200
# logging ip access-list cache threshold 5000
# logging severity auth 2
# logging severity ospfv3 1
# logging severity ftp 6
# logging origin-id hostname
# logging server 203.0.113.100 1 use-vrf management
# logging server 203.0.113.101 3 use-vrf default facility local6

- name: Replace logging configurations with provided config
  cisco.nxos.nxos_logging_global:
    config:
      monitor:
        severity: warning
      ip:
        access_list:
          cache:
            entries: 4096
      facilities:
        - facility: auth
          severity: critical
        - facility: ospfv3
          severity: alert
        - facility: ftp
          severity: informational
      hosts:
        - host: 203.0.113.101
          severity: error
          facility: local6
          use_vrf: default
        - host: 198.51.100.101
          severity: alert
          port: 6538
          use_vrf: management
      origin_id:
        ip: 192.0.2.100
    state: replaced

# Task output
# -------------
# before:
#   console:
#      severity: error
#    facilities:
#      - facility: auth
#        severity: critical
#      - facility: ftp
#        severity: informational
#      - facility: ospfv3
#        severity: alert
#    ip:
#      access_list:
#        cache:
#          entries: 16384
#          interval: 200
#          threshold: 5000
#    monitor:
#      severity: warning
#    origin_id:
#      hostname: true
#    hosts:
#      - severity: alert
#        host: 203.0.113.100
#        use_vrf: management
#      - facility: local6
#        severity: error
#        host: 203.0.113.101
#        use_vrf: default
#
# commands:
#   - "logging console"
#   - "logging ip access-list cache entries 4096"
#   - "no logging ip access-list cache interval 200"
#   - "no logging ip access-list cache threshold 5000"
#   - "no logging origin-id hostname"
#   - "logging origin-id ip 192.0.2.100"
#   - "logging server 198.51.100.101 1 port 6538 use-vrf management"
#   - "no logging server 203.0.113.100 1 use-vrf management"
#
#  after:
#    facilities:
#      - facility: auth
#        severity: critical
#      - facility: ftp
#        severity: informational
#      - facility: ospfv3
#        severity: alert
#    ip:
#      access_list:
#        cache:
#          entries: 4096
#    monitor:
#      severity: warning
#    origin_id:
#      ip: 192.0.2.100
#    hosts:
#      - severity: alert
#        port: 6538
#        host: 198.51.100.101
#        use_vrf: management
#      - facility: local6
#        severity: error
#        host: 203.0.113.101
#        use_vrf: default
#
# After state:
# ------------
# nxos-9k-rdo# show running-config | include logging
# logging monitor 4
# logging ip access-list cache entries 4096
# logging severity auth 2
# logging severity ospfv3 1
# logging severity ftp 6
# logging origin-id ip 192.0.2.100
# logging server 203.0.113.101 3 use-vrf default facility local6
# logging server 198.51.100.101 1 port 6538 use-vrf management

# Using deleted to delete all logging configurations

# Before state:
# ------------
# nxos-9k-rdo# show running-config | include logging
# logging console 3
# logging monitor 4
# logging ip access-list cache entries 16384
# logging ip access-list cache interval 200
# logging ip access-list cache threshold 5000
# logging severity auth 2
# logging severity ospfv3 1
# logging severity ftp 6
# logging origin-id hostname
# logging server 203.0.113.100 1 use-vrf management
# logging server 203.0.113.101 3 use-vrf default facility local6

- name: Delete all logging configuration
  cisco.nxos.nxos_logging_global:
    state: deleted

# Task output
# -------------
# before:
#   console:
#      severity: error
#    facilities:
#      - facility: auth
#        severity: critical
#      - facility: ftp
#        severity: informational
#      - facility: ospfv3
#        severity: alert
#    ip:
#      access_list:
#        cache:
#          entries: 16384
#          interval: 200
#          threshold: 5000
#    monitor:
#      severity: warning
#    origin_id:
#      hostname: true
#    hosts:
#      - severity: alert
#        host: 203.0.113.100
#        use_vrf: management
#      - facility: local6
#        severity: error
#        host: 203.0.113.101
#        use_vrf: default
#
# commands:
#   - "logging console"
#   - "logging monitor"
#   - "no logging ip access-list cache entries 16384"
#   - "no logging ip access-list cache interval 200"
#   - "no logging ip access-list cache threshold 5000"
#   - "no logging origin-id hostname"
#   - "no logging severity auth 2"
#   - "no logging severity ospfv3 1"
#   - "no logging severity ftp 6"
#   - "no logging server 203.0.113.100 1 use-vrf management"
#   - "no logging server 203.0.113.101 3 facility local6 use-vrf default"
#
# after: {}

# Using rendered

- name: Render platform specific configuration lines with state rendered (without connecting to the device)
  cisco.nxos.nxos_logging_global:
    config:
      console:
        severity: error
      monitor:
        severity: warning
      ip:
        access_list:
          cache:
            entries: 16384
            interval: 200
            threshold: 5000
      facilities:
        - facility: auth
          severity: critical
        - facility: ospfv3
          severity: alert
        - facility: ftp
          severity: informational
      hosts:
        - host: 203.0.113.100
          severity: alert
          use_vrf: management
        - host: 203.0.113.101
          severity: error
          facility: local6
          use_vrf: default
      origin_id:
        hostname: True

# Task Output (redacted)
# -----------------------
#  rendered:
#    - "logging console 3"
#    - "logging monitor 4"
#    - "logging ip access-list cache entries 16384"
#    - "logging ip access-list cache interval 200"
#    - "logging ip access-list cache threshold 5000"
#    - "logging severity auth 2"
#    - "logging severity ospfv3 1"
#    - "logging severity ftp 6"
#    - "logging server 203.0.113.100 1 use-vrf management"
#    - "logging server 203.0.113.101 3 facility local6 use-vrf default"
#    - "logging origin-id hostname"

# Using parsed

# parsed.cfg
# ------------
# logging console 3
# logging monitor 4
# logging ip access-list cache entries 16384
# logging ip access-list cache interval 200
# logging ip access-list cache threshold 5000
# logging severity auth 2
# logging severity ospfv3 1
# logging severity ftp 6
# logging origin-id hostname
# logging server 203.0.113.100 1 use-vrf management
# logging server 203.0.113.101 3 use-vrf default facility local6

- name: Parse externally provided logging configuration
  cisco.nxos.nxos_logging_global:
    running_config: "{{ lookup('file', './fixtures/parsed.cfg') }}"
    state: parsed

# Task output (redacted)
# -----------------------
# parsed:
#   console:
#      severity: error
#    facilities:
#      - facility: auth
#        severity: critical
#      - facility: ftp
#        severity: informational
#      - facility: ospfv3
#        severity: alert
#    ip:
#      access_list:
#        cache:
#          entries: 16384
#          interval: 200
#          threshold: 5000
#    monitor:
#      severity: warning
#    origin_id:
#      hostname: true
#    hosts:
#      - severity: alert
#        host: 203.0.113.100
#        use_vrf: management
#      - facility: local6
#        severity: error
#        host: 203.0.113.101
#        use_vrf: default
"""

RETURN = """
before:
  description: The configuration prior to the module execution.
  returned: when state is I(merged), I(replaced), I(overridden), I(deleted) or I(purged)
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
after:
  description: The resulting configuration after module execution.
  returned: when changed
  type: dict
  sample: >
    This output will always be in the same format as the
    module argspec.
commands:
  description: The set of commands pushed to the remote device.
  returned: when state is I(merged), I(replaced), I(overridden), I(deleted) or I(purged)
  type: list
  sample:
    - "logging console 3"
    - "logging monitor 4"
    - "logging ip access-list cache entries 16384"
    - "logging ip access-list cache interval 200"
    - "logging ip access-list cache threshold 5000"
rendered:
  description: The provided configuration in the task rendered in device-native format (offline).
  returned: when state is I(rendered)
  type: list
  sample:
    - "logging ip access-list cache entries 4096"
    - "no logging ip access-list cache interval 200"
    - "no logging ip access-list cache threshold 5000"
    - "no logging origin-id hostname"
    - "logging origin-id ip 192.0.2.100"
    - "logging server 198.51.100.101 1 port 6538 use-vrf management"
gathered:
  description: Facts about the network resource gathered from the remote device as structured data.
  returned: when state is I(gathered)
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
parsed:
  description: The device native config provided in I(running_config) option parsed into structured data as per module argspec.
  returned: when state is I(parsed)
  type: list
  sample: >
    This output will always be in the same format as the
    module argspec.
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.logging_global.logging_global import (
    Logging_globalArgs,
)
from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.config.logging_global.logging_global import (
    Logging_global,
)


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    module = AnsibleModule(
        argument_spec=Logging_globalArgs.argument_spec,
        mutually_exclusive=[["config", "running_config"]],
        required_if=[
            ["state", "merged", ["config"]],
            ["state", "replaced", ["config"]],
            ["state", "overridden", ["config"]],
            ["state", "rendered", ["config"]],
            ["state", "parsed", ["running_config"]],
        ],
        supports_check_mode=True,
    )

    result = Logging_global(module).execute_module()
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type
from ansible.plugins.action import ActionBase
try:
    from ansible_collections.ansible.utils.plugins.module_utils.common.argspec_validate import (
        AnsibleArgSpecValidator,
    )
except ImportError:
    ANSIBLE_UTILS_IS_INSTALLED = False
else:
    ANSIBLE_UTILS_IS_INSTALLED = True
from ansible.errors import AnsibleActionFail
from ansible_collections.cisco.dnac.plugins.plugin_utils.dnac import (
    DNACSDK,
    dnac_argument_spec,
    dnac_compare_equality2,
    get_dict_result,
)
from ansible_collections.cisco.dnac.plugins.plugin_utils.exceptions import (
    InconsistentParameters,
)

# Get common arguments specification
argument_spec = dnac_argument_spec()
# Add arguments specific for this module
argument_spec.update(dict(
    state=dict(type="str", default="present", choices=["present", "absent"]),
    cliTransport=dict(type="str"),
    computeDevice=dict(type="bool"),
    enablePassword=dict(type="str", no_log=True),
    extendedDiscoveryInfo=dict(type="str"),
    httpPassword=dict(type="str"),
    httpPort=dict(type="str"),
    httpSecure=dict(type="bool"),
    httpUserName=dict(type="str"),
    ipAddress=dict(type="list"),
    merakiOrgId=dict(type="list"),
    netconfPort=dict(type="str"),
    password=dict(type="str", no_log=True),
    serialNumber=dict(type="str"),
    snmpAuthPassphrase=dict(type="str"),
    snmpAuthProtocol=dict(type="str"),
    snmpMode=dict(type="str"),
    snmpPrivPassphrase=dict(type="str"),
    snmpPrivProtocol=dict(type="str"),
    snmpROCommunity=dict(type="str"),
    snmpRWCommunity=dict(type="str"),
    snmpRetry=dict(type="int"),
    snmpTimeout=dict(type="int"),
    snmpUserName=dict(type="str"),
    snmpVersion=dict(type="str"),
    type=dict(type="str"),
    updateMgmtIPaddressList=dict(type="list"),
    userName=dict(type="str"),
    id=dict(type="str"),
    cleanConfig=dict(type="bool"),
))

required_if = [
    ("state", "present", ["id", "ipAddress"], True),
    ("state", "absent", ["id", "ipAddress"], True),
]
required_one_of = []
mutually_exclusive = []
required_together = []


class NetworkDevice(object):
    def __init__(self, params, dnac):
        self.dnac = dnac
        self.new_object = dict(
            cliTransport=params.get("cliTransport"),
            computeDevice=params.get("computeDevice"),
            enablePassword=params.get("enablePassword"),
            extendedDiscoveryInfo=params.get("extendedDiscoveryInfo"),
            httpPassword=params.get("httpPassword"),
            httpPort=params.get("httpPort"),
            httpSecure=params.get("httpSecure"),
            httpUserName=params.get("httpUserName"),
            ipAddress=params.get("ipAddress"),
            merakiOrgId=params.get("merakiOrgId"),
            netconfPort=params.get("netconfPort"),
            password=params.get("password"),
            serialNumber=params.get("serialNumber"),
            snmpAuthPassphrase=params.get("snmpAuthPassphrase"),
            snmpAuthProtocol=params.get("snmpAuthProtocol"),
            snmpMode=params.get("snmpMode"),
            snmpPrivPassphrase=params.get("snmpPrivPassphrase"),
            snmpPrivProtocol=params.get("snmpPrivProtocol"),
            snmpROCommunity=params.get("snmpROCommunity"),
            snmpRWCommunity=params.get("snmpRWCommunity"),
            snmpRetry=params.get("snmpRetry"),
            snmpTimeout=params.get("snmpTimeout"),
            snmpUserName=params.get("snmpUserName"),
            snmpVersion=params.get("snmpVersion"),
            type=params.get("type"),
            updateMgmtIPaddressList=params.get("updateMgmtIPaddressList"),
            userName=params.get("userName"),
            id=params.get("id"),
            clean_config=params.get("cleanConfig"),
            managementIpAddress=params.get("managementIpAddress"),
        )

    def get_all_params(self, name=None, id=None):
        new_object_params = {}
        new_object_params['hostname'] = self.new_object.get('hostname')
        new_object_params['management_ip_address'] = self.new_object.get('management_ip_address') or \
            self.new_object.get('ipAddress')
        new_object_params['mac_address'] = self.new_object.get('macAddress') or \
            self.new_object.get('mac_address')
        new_object_params['location_name'] = self.new_object.get('locationName') or \
            self.new_object.get('location_name')
        new_object_params['serial_number'] = self.new_object.get('serialNumber') or \
            self.new_object.get('serial_number')
        new_object_params['location'] = self.new_object.get('location')
        new_object_params['family'] = self.new_object.get('family')
        # new_object_params['type'] = self.new_object.get('type')
        new_object_params['series'] = self.new_object.get('series')
        new_object_params['collection_status'] = self.new_object.get('collectionStatus') or \
            self.new_object.get('collection_status')
        new_object_params['collection_interval'] = self.new_object.get('collectionInterval') or \
            self.new_object.get('collection_interval')
        new_object_params['not_synced_for_minutes'] = self.new_object.get('notSyncedForMinutes') or \
            self.new_object.get('not_synced_for_minutes')
        new_object_params['error_code'] = self.new_object.get('errorCode') or \
            self.new_object.get('error_code')
        new_object_params['error_description'] = self.new_object.get('errorDescription') or \
            self.new_object.get('error_description')
        new_object_params['software_version'] = self.new_object.get('softwareVersion') or \
            self.new_object.get('software_version')
        new_object_params['software_type'] = self.new_object.get('softwareType') or \
            self.new_object.get('software_type')
        new_object_params['platform_id'] = self.new_object.get('platformId') or \
            self.new_object.get('platform_id')
        new_object_params['role'] = self.new_object.get('role')
        new_object_params['reachability_status'] = self.new_object.get('reachabilityStatus') or \
            self.new_object.get('reachability_status')
        new_object_params['up_time'] = self.new_object.get('upTime') or \
            self.new_object.get('up_time')
        new_object_params['associated_wlc_ip'] = self.new_object.get('associatedWlcIp') or \
            self.new_object.get('associated_wlc_ip')
        new_object_params['license_name'] = self.new_object.get('license_name')
        new_object_params['license_type'] = self.new_object.get('license_type')
        new_object_params['license_status'] = self.new_object.get('license_status')
        new_object_params['module_name'] = self.new_object.get('module_name')
        new_object_params['module_equpimenttype'] = self.new_object.get('module_equpimenttype')
        new_object_params['module_servicestate'] = self.new_object.get('module_servicestate')
        new_object_params['module_vendorequipmenttype'] = self.new_object.get('module_vendorequipmenttype')
        new_object_params['module_partnumber'] = self.new_object.get('module_partnumber')
        new_object_params['module_operationstatecode'] = self.new_object.get('module_operationstatecode')
        new_object_params['id'] = id or self.new_object.get('id')
        new_object_params['device_support_level'] = self.new_object.get('deviceSupportLevel') or \
            self.new_object.get('device_support_level')
        new_object_params['offset'] = self.new_object.get('offset')
        new_object_params['limit'] = self.new_object.get('limit')
        return new_object_params

    def create_params(self):
        new_object_params = {}
        new_object_params['cliTransport'] = self.new_object.get('cliTransport')
        new_object_params['computeDevice'] = self.new_object.get('computeDevice')
        new_object_params['enablePassword'] = self.new_object.get('enablePassword')
        new_object_params['extendedDiscoveryInfo'] = self.new_object.get('extendedDiscoveryInfo')
        new_object_params['httpPassword'] = self.new_object.get('httpPassword')
        new_object_params['httpPort'] = self.new_object.get('httpPort')
        new_object_params['httpSecure'] = self.new_object.get('httpSecure')
        new_object_params['httpUserName'] = self.new_object.get('httpUserName')
        new_object_params['ipAddress'] = self.new_object.get('ipAddress')
        new_object_params['merakiOrgId'] = self.new_object.get('merakiOrgId')
        new_object_params['netconfPort'] = self.new_object.get('netconfPort')
        new_object_params['password'] = self.new_object.get('password')
        new_object_params['serialNumber'] = self.new_object.get('serialNumber')
        new_object_params['snmpAuthPassphrase'] = self.new_object.get('snmpAuthPassphrase')
        new_object_params['snmpAuthProtocol'] = self.new_object.get('snmpAuthProtocol')
        new_object_params['snmpMode'] = self.new_object.get('snmpMode')
        new_object_params['snmpPrivPassphrase'] = self.new_object.get('snmpPrivPassphrase')
        new_object_params['snmpPrivProtocol'] = self.new_object.get('snmpPrivProtocol')
        new_object_params['snmpROCommunity'] = self.new_object.get('snmpROCommunity')
        new_object_params['snmpRWCommunity'] = self.new_object.get('snmpRWCommunity')
        new_object_params['snmpRetry'] = self.new_object.get('snmpRetry')
        new_object_params['snmpTimeout'] = self.new_object.get('snmpTimeout')
        new_object_params['snmpUserName'] = self.new_object.get('snmpUserName')
        new_object_params['snmpVersion'] = self.new_object.get('snmpVersion')
        new_object_params['type'] = self.new_object.get('type')
        new_object_params['updateMgmtIPaddressList'] = self.new_object.get('updateMgmtIPaddressList')
        new_object_params['userName'] = self.new_object.get('userName')
        return new_object_params

    def delete_by_id_params(self):
        new_object_params = {}
        new_object_params['clean_config'] = self.new_object.get('clean_config')
        new_object_params['id'] = self.new_object.get('id')
        return new_object_params

    def update_all_params(self):
        new_object_params = {}
        new_object_params['cliTransport'] = self.new_object.get('cliTransport')
        new_object_params['computeDevice'] = self.new_object.get('computeDevice')
        new_object_params['enablePassword'] = self.new_object.get('enablePassword')
        new_object_params['extendedDiscoveryInfo'] = self.new_object.get('extendedDiscoveryInfo')
        new_object_params['httpPassword'] = self.new_object.get('httpPassword')
        new_object_params['httpPort'] = self.new_object.get('httpPort')
        new_object_params['httpSecure'] = self.new_object.get('httpSecure')
        new_object_params['httpUserName'] = self.new_object.get('httpUserName')
        new_object_params['ipAddress'] = self.new_object.get('ipAddress')
        new_object_params['merakiOrgId'] = self.new_object.get('merakiOrgId')
        new_object_params['netconfPort'] = self.new_object.get('netconfPort')
        new_object_params['password'] = self.new_object.get('password')
        new_object_params['serialNumber'] = self.new_object.get('serialNumber')
        new_object_params['snmpAuthPassphrase'] = self.new_object.get('snmpAuthPassphrase')
        new_object_params['snmpAuthProtocol'] = self.new_object.get('snmpAuthProtocol')
        new_object_params['snmpMode'] = self.new_object.get('snmpMode')
        new_object_params['snmpPrivPassphrase'] = self.new_object.get('snmpPrivPassphrase')
        new_object_params['snmpPrivProtocol'] = self.new_object.get('snmpPrivProtocol')
        new_object_params['snmpROCommunity'] = self.new_object.get('snmpROCommunity')
        new_object_params['snmpRWCommunity'] = self.new_object.get('snmpRWCommunity')
        new_object_params['snmpRetry'] = self.new_object.get('snmpRetry')
        new_object_params['snmpTimeout'] = self.new_object.get('snmpTimeout')
        new_object_params['snmpUserName'] = self.new_object.get('snmpUserName')
        new_object_params['snmpVersion'] = self.new_object.get('snmpVersion')
        new_object_params['type'] = self.new_object.get('type')
        new_object_params['updateMgmtIPaddressList'] = self.new_object.get('updateMgmtIPaddressList')
        new_object_params['userName'] = self.new_object.get('userName')
        return new_object_params

    def get_object_by_name(self, name):
        result = None
        # NOTE: Does not have a get by name method or it is in another action
        try:
            items = self.dnac.exec(
                family="devices",
                function="get_device_list",
                params=self.get_all_params(name=name),
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = get_dict_result(items, 'name', name)
        except Exception:
            result = None
        return result

    def get_object_by_id(self, id):
        result = None
        try:
            items = self.dnac.exec(
                family="devices",
                function="get_device_by_id",
                params={"id": id}
            )
            if isinstance(items, dict):
                if 'response' in items:
                    items = items.get('response')
            result = get_dict_result(items, 'id', id)
        except Exception:
            result = None
        return result

    def exists(self):
        id_exists = False
        name_exists = False
        prev_obj = None
        o_id = self.new_object.get("id")
        name = self.new_object.get("name") or \
            self.new_object.get('ipAddress')
        if isinstance(name, list) and len(name) > 0:
            name = name[0]
        if o_id:
            prev_obj = self.get_object_by_id(o_id)
            id_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if not id_exists and name:
            prev_obj = self.get_object_by_name(name)
            name_exists = prev_obj is not None and isinstance(prev_obj, dict)
        if name_exists:
            _id = prev_obj.get("id")
            if id_exists and name_exists and o_id != _id:
                raise InconsistentParameters("The 'id' and 'name' params don't refer to the same object")
            if _id:
                self.new_object.update(dict(id=_id))
            if _id:
                prev_obj = self.get_object_by_id(_id)
        it_exists = prev_obj is not None and isinstance(prev_obj, dict)
        return (it_exists, prev_obj)

    def requires_update(self, current_obj):
        requested_obj = self.new_object

        obj_params = [
            ("cliTransport", "cliTransport"),
            ("computeDevice", "computeDevice"),
            ("enablePassword", "enablePassword"),
            ("extendedDiscoveryInfo", "extendedDiscoveryInfo"),
            ("httpPassword", "httpPassword"),
            ("httpPort", "httpPort"),
            ("httpSecure", "httpSecure"),
            ("httpUserName", "httpUserName"),
            ("ipAddress", "ipAddress"),
            ("merakiOrgId", "merakiOrgId"),
            ("netconfPort", "netconfPort"),
            ("serialNumber", "serialNumber"),
            ("snmpAuthPassphrase", "snmpAuthPassphrase"),
            ("snmpAuthProtocol", "snmpAuthProtocol"),
            ("snmpMode", "snmpMode"),
            ("snmpPrivPassphrase", "snmpPrivPassphrase"),
            ("snmpPrivProtocol", "snmpPrivProtocol"),
            ("snmpROCommunity", "snmpROCommunity"),
            ("snmpRWCommunity", "snmpRWCommunity"),
            ("snmpRetry", "snmpRetry"),
            ("snmpTimeout", "snmpTimeout"),
            ("snmpUserName", "snmpUserName"),
            ("snmpVersion", "snmpVersion"),
            ("type", "type"),
            ("updateMgmtIPaddressList", "updateMgmtIPaddressList"),
            ("userName", "userName"),
            ("id", "id"),
            ("cleanConfig", "clean_config"),
        ]
        # Method 1. Params present in request (Ansible) obj are the same as the current (DNAC) params
        # If any does not have eq params, it requires update
        return any(not dnac_compare_equality2(current_obj.get(dnac_param),
                                              requested_obj.get(ansible_param))
                   for (dnac_param, ansible_param) in obj_params)

    def create(self):
        result = self.dnac.exec(
            family="devices",
            function="add_device",
            params=self.create_params(),
            op_modifies=True,
        )
        return result

    def update(self):
        id = self.new_object.get("id")
        name = self.new_object.get("name") or \
            self.new_object.get('ipAddress')
        if isinstance(name, list) and len(name) > 0:
            name = name[0]
        result = None
        result = self.dnac.exec(
            family="devices",
            function="sync_devices",
            params=self.update_all_params(),
            op_modifies=True,
        )
        return result

    def delete(self):
        id = self.new_object.get("id")
        name = self.new_object.get("name") or \
            self.new_object.get('ipAddress')
        if isinstance(name, list) and len(name) > 0:
            name = name[0]
        result = None
        if not id:
            prev_obj_name = self.get_object_by_name(name)
            id_ = None
            if prev_obj_name:
                id_ = prev_obj_name.get("id")
            if id_:
                self.new_object.update(dict(id=id_))
        result = self.dnac.exec(
            family="devices",
            function="delete_device_by_id",
            params=self.delete_by_id_params(),
        )
        return result


class ActionModule(ActionBase):
    def __init__(self, *args, **kwargs):
        if not ANSIBLE_UTILS_IS_INSTALLED:
            raise AnsibleActionFail("ansible.utils is not installed. Execute 'ansible-galaxy collection install ansible.utils'")
        super(ActionModule, self).__init__(*args, **kwargs)
        self._supports_async = False
        self._supports_check_mode = False
        self._result = None

    # Checks the supplied parameters against the argument spec for this module
    def _check_argspec(self):
        aav = AnsibleArgSpecValidator(
            data=self._task.args,
            schema=dict(argument_spec=argument_spec),
            schema_format="argspec",
            schema_conditionals=dict(
                required_if=required_if,
                required_one_of=required_one_of,
                mutually_exclusive=mutually_exclusive,
                required_together=required_together,
            ),
            name=self._task.action,
        )
        valid, errors, self._task.args = aav.validate()
        if not valid:
            raise AnsibleActionFail(errors)

    def run(self, tmp=None, task_vars=None):
        self._task.diff = False
        self._result = super(ActionModule, self).run(tmp, task_vars)
        self._result["changed"] = False
        self._check_argspec()

        dnac = DNACSDK(self._task.args)
        obj = NetworkDevice(self._task.args, dnac)

        state = self._task.args.get("state")

        response = None

        if state == "present":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                if obj.requires_update(prev_obj):
                    response = obj.update()
                    dnac.object_updated()
                else:
                    response = prev_obj
                    dnac.object_already_present()
            else:
                response = obj.create()
                dnac.object_created()

        elif state == "absent":
            (obj_exists, prev_obj) = obj.exists()
            if obj_exists:
                response = obj.delete()
                dnac.object_deleted()
            else:
                dnac.object_already_absent()

        self._result.update(dict(dnac_response=response))
        self._result.update(dnac.exit_json())
        return self._result



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: sda_fabric_site
short_description: Resource module for Sda Fabric Site
description:
- Manage operations create and delete of the resource Sda Fabric Site.
- Add Site in SDA Fabric.
- Delete Site from SDA Fabric.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  fabricName:
    description: Warning - Starting DNA Center 2.2.3.5 release, this field has been
      deprecated. SD-Access Fabric does not need it anymore. It will be removed in future
      DNA Center releases.
    type: str
    version_added: 4.0.0
  fabricType:
    description: Type of SD-Access Fabric. Allowed values are "FABRIC_SITE" or "FABRIC_ZONE".
      Default value is "FABRIC_SITE".
    type: str
  siteNameHierarchy:
    description: SiteNameHierarchy query parameter. Site Name Hierarchy.
    type: str
requirements:
- dnacentersdk >= 2.5.5
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for SDA AddSiteInSDAFabric
  description: Complete reference of the AddSiteInSDAFabric API.
  link: https://developer.cisco.com/docs/dna-center/#!add-site-in-sda-fabric
- name: Cisco DNA Center documentation for SDA DeleteSiteFromSDAFabric
  description: Complete reference of the DeleteSiteFromSDAFabric API.
  link: https://developer.cisco.com/docs/dna-center/#!delete-site-from-sda-fabric
notes:
  - SDK Method used are
    sda.Sda.add_site,
    sda.Sda.delete_site,

  - Paths used are
    post /dna/intent/api/v1/business/sda/fabric-site,
    delete /dna/intent/api/v1/business/sda/fabric-site,

"""

EXAMPLES = r"""
- name: Delete all
  cisco.dnac.sda_fabric_site:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: absent
    siteNameHierarchy: string

- name: Create
  cisco.dnac.sda_fabric_site:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    fabricName: string
    fabricType: string
    siteNameHierarchy: string

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "status": "string",
      "description": "string",
      "taskId": "string",
      "taskStatusUrl": "string",
      "executionStatusUrl": "string",
      "executionId": "string"
    }
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: applications
short_description: Resource module for Applications
description:
- Manage operations create, update and delete of the resource Applications.
- Create new Custom application.
- Delete existing application by its id.
- Edit the attributes of an existing application.
version_added: '3.1.0'
extends_documentation_fragment:
  - cisco.dnac.module
author: Rafael Campos (@racampos)
options:
  id:
    description: Id query parameter. Application's Id.
    type: str
  payload:
    description: Applications's payload.
    elements: dict
    suboptions:
      applicationSet:
        description: Applications's applicationSet.
        suboptions:
          idRef:
            description: Id Ref.
            type: str
        type: dict
      indicativeNetworkIdentity:
        description: Applications's indicativeNetworkIdentity.
        elements: dict
        suboptions:
          displayName:
            description: DisplayName.
            type: str
          id:
            description: Id.
            type: str
          lowerPort:
            description: LowerPort.
            type: int
          ports:
            description: Ports.
            type: str
          protocol:
            description: Protocol.
            type: str
          upperPort:
            description: UpperPort.
            type: int
        type: list
      name:
        description: Name.
        type: str
      networkApplications:
        description: Applications's networkApplications.
        elements: dict
        suboptions:
          appProtocol:
            description: App Protocol.
            type: str
          applicationSubType:
            description: Application Sub Type.
            type: str
          applicationType:
            description: Application Type.
            type: str
          categoryId:
            description: Category Id.
            type: str
          displayName:
            description: Display Name.
            type: str
          dscp:
            description: Dscp.
            type: str
          engineId:
            description: Engine Id.
            type: str
          helpString:
            description: Help String.
            type: str
          ignoreConflict:
            description: Ignore Conflict.
            type: str
          longDescription:
            description: Long Description.
            type: str
          name:
            description: Name.
            type: str
          popularity:
            description: Popularity.
            type: str
          rank:
            description: Rank.
            type: str
          serverName:
            description: Server Name.
            type: str
          trafficClass:
            description: Traffic Class.
            type: str
          url:
            description: Url.
            type: str
        type: list
      networkIdentity:
        description: Applications's networkIdentity.
        elements: dict
        suboptions:
          displayName:
            description: Display Name.
            type: str
          lowerPort:
            description: Lower Port.
            type: str
          ports:
            description: Ports.
            type: str
          protocol:
            description: Protocol.
            type: str
          upperPort:
            description: Upper Port.
            type: str
        type: list
    type: list
requirements:
- dnacentersdk >= 2.5.5
- python >= 3.5
seealso:
- name: Cisco DNA Center documentation for Application Policy CreateApplication
  description: Complete reference of the CreateApplication API.
  link: https://developer.cisco.com/docs/dna-center/#!create-application
- name: Cisco DNA Center documentation for Application Policy DeleteApplication
  description: Complete reference of the DeleteApplication API.
  link: https://developer.cisco.com/docs/dna-center/#!delete-application
- name: Cisco DNA Center documentation for Application Policy EditApplication
  description: Complete reference of the EditApplication API.
  link: https://developer.cisco.com/docs/dna-center/#!edit-application
notes:
  - SDK Method used are
    application_policy.ApplicationPolicy.create_application,
    application_policy.ApplicationPolicy.delete_application,
    application_policy.ApplicationPolicy.edit_application,

  - Paths used are
    post /dna/intent/api/v1/applications,
    delete /dna/intent/api/v1/applications,
    put /dna/intent/api/v1/applications,

"""

EXAMPLES = r"""
- name: Create
  cisco.dnac.applications:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    payload:
    - applicationSet:
        idRef: string
      indicativeNetworkIdentity:
      - displayName: string
        id: string
        lowerPort: 0
        ports: string
        protocol: string
        upperPort: 0
      name: string
      networkApplications:
      - appProtocol: string
        applicationSubType: string
        applicationType: string
        categoryId: string
        displayName: string
        dscp: string
        engineId: string
        helpString: string
        ignoreConflict: string
        longDescription: string
        name: string
        popularity: string
        rank: string
        serverName: string
        trafficClass: string
        url: string
      networkIdentity:
      - displayName: string
        lowerPort: string
        ports: string
        protocol: string
        upperPort: string

- name: Update all
  cisco.dnac.applications:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: present
    payload:
    - applicationSet:
        idRef: string
      id: string
      name: string
      networkApplications:
      - appProtocol: string
        applicationSubType: string
        applicationType: string
        categoryId: string
        displayName: string
        dscp: string
        engineId: string
        helpString: string
        id: string
        ignoreConflict: string
        longDescription: string
        name: string
        popularity: string
        rank: string
        serverName: string
        trafficClass: string
        url: string
      networkIdentity:
      - displayName: string
        id: string
        lowerPort: string
        ports: string
        protocol: string
        upperPort: string

- name: Delete all
  cisco.dnac.applications:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    state: absent
    id: string

"""

RETURN = r"""
dnac_response:
  description: A dictionary or list with the response returned by the Cisco DNAC Python SDK
  returned: always
  type: dict
  sample: >
    {
      "taskId": "string",
      "url": "string"
    }
"""



# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat Inc.
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
"""
The iosxr_l2_interfaces class
It is in this file where the current configuration (as dict)
is compared to the provided configuration (as dict) and the command set
necessary to bring the current configuration to it's desired end-state is
created
"""

from __future__ import absolute_import, division, print_function


__metaclass__ = type

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.cfg.base import (
    ConfigBase,
)
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    remove_empties,
    to_list,
)

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.facts.facts import Facts
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.iosxr import get_os_version
from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.utils.utils import (
    Version,
    add_command_to_config_list,
    dict_diff,
    filter_dict_having_none_value,
    normalize_interface,
    remove_command_from_config_list,
    remove_duplicate_interface,
)


class L2_Interfaces(ConfigBase):
    """
    The iosxr_interfaces class
    """

    gather_subset = ["!all", "!min"]

    gather_network_resources = ["l2_interfaces"]

    def get_l2_interfaces_facts(self, data=None):
        """Get the 'facts' (the current configuration)
        :rtype: A dictionary
        :returns: The current configuration as a dictionary
        """
        facts, _warnings = Facts(self._module).get_facts(
            self.gather_subset,
            self.gather_network_resources,
            data=data,
        )
        l2_interfaces_facts = facts["ansible_network_resources"].get(
            "l2_interfaces",
        )
        if not l2_interfaces_facts:
            return []
        return l2_interfaces_facts

    def execute_module(self):
        """Execute the module
        :rtype: A dictionary
        :returns: The result from module execution
        """
        result = {"changed": False}
        warnings = list()
        commands = list()

        if self.state in self.ACTION_STATES:
            existing_l2_interfaces_facts = self.get_l2_interfaces_facts()
        else:
            existing_l2_interfaces_facts = []

        if self.state in self.ACTION_STATES or self.state == "rendered":
            commands.extend(self.set_config(existing_l2_interfaces_facts))

        if commands and self.state in self.ACTION_STATES:
            if not self._module.check_mode:
                self._connection.edit_config(commands)
            result["changed"] = True

        if self.state in self.ACTION_STATES:
            result["commands"] = commands

        if self.state in self.ACTION_STATES or self.state == "gathered":
            changed_l2_interfaces_facts = self.get_l2_interfaces_facts()

        elif self.state == "rendered":
            result["rendered"] = commands

        elif self.state == "parsed":
            running_config = self._module.params["running_config"]
            if not running_config:
                self._module.fail_json(
                    msg="value of running_config parameter must not be empty for state parsed",
                )
            result["parsed"] = self.get_l2_interfaces_facts(
                data=running_config,
            )

        if self.state in self.ACTION_STATES:
            result["before"] = existing_l2_interfaces_facts
            if result["changed"]:
                result["after"] = changed_l2_interfaces_facts

        elif self.state == "gathered":
            result["gathered"] = changed_l2_interfaces_facts

        result["warnings"] = warnings
        return result

    def set_config(self, existing_l2_interfaces_facts):
        """Collect the configuration from the args passed to the module,
            collect the current configuration (as a dict from facts)
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        want = self._module.params["config"]
        have = existing_l2_interfaces_facts
        resp = self.set_state(want, have)
        return to_list(resp)

    def set_state(self, want, have):
        """Select the appropriate function based on the state provided
        :param want: the desired configuration as a dictionary
        :param have: the current configuration as a dictionary
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []

        if self.state in ("overridden", "merged", "replaced", "rendered") and not want:
            self._module.fail_json(
                msg="value of config parameter must not be empty for state {0}".format(
                    self.state,
                ),
            )

        if self.state == "overridden":
            commands = self._state_overridden(want, have, self._module)
        elif self.state == "deleted":
            commands = self._state_deleted(want, have)
        elif self.state in ("merged", "rendered"):
            commands = self._state_merged(want, have, self._module)
        elif self.state == "replaced":
            commands = self._state_replaced(want, have, self._module)

        return commands

    def _state_replaced(self, want, have, module):
        """The command generator when state is replaced
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []
        for interface in want:
            interface["name"] = normalize_interface(interface["name"])
            for each in have:
                if each["name"] == interface["name"] or interface["name"] in each["name"]:
                    break
            else:
                commands.extend(self._set_config(interface, {}, module))
                continue
            interface = remove_empties(interface)
            have_dict = filter_dict_having_none_value(interface, each)
            commands.extend(self._clear_config(dict(), have_dict))
            commands.extend(self._set_config(interface, each, module))
        # Remove the duplicate interface call
        commands = remove_duplicate_interface(commands)

        return commands

    def _state_overridden(self, want, have, module):
        """The command generator when state is overridden
        :rtype: A list
        :returns: the commands necessary to migrate the current configuration
                  to the desired configuration
        """
        commands = []
        not_in_have = set()
        in_have = set()
        for each in have:
            for interface in want:
                interface["name"] = normalize_interface(interface["name"])
                if each["name"] == interface["name"]:
                    in_have.add(interface["name"])
                    break
                if interface["name"] != each["name"]:
                    not_in_have.add(interface["name"])
            else:
                # We didn't find a matching desired state, which means we can
                # pretend we received an empty desired state.
                interface = dict(name=each["name"])
                commands.extend(self._clear_config(interface, each))
                continue
            interface = remove_empties(interface)
            have_dict = filter_dict_having_none_value(interface, each)
            commands.extend(self._clear_config(dict(), have_dict))
            commands.extend(self._set_config(interface, each, module))
        # Add the want interface that's not already configured in have interface
        for each in not_in_have - in_have:
            for every in want:
                interface = "interface {0}".format(every["name"])
                if each and interface not in commands:
                    commands.extend(self._set_config(every, {}, module))

        # Remove the duplicate interface call
        commands = remove_duplicate_interface(commands)

        return commands

    def _state_merged(self, want, have, module):
        """The command generator when state is merged
        :rtype: A list
        :returns: the commands necessary to merge the provided into
                  the current configuration
        """
        commands = []

        for interface in want:
            interface["name"] = normalize_interface(interface["name"])
            interface = remove_empties(interface)
            for each in have:
                if each["name"] == interface["name"] or interface["name"] in each["name"]:
                    break
            else:
                commands.extend(self._set_config(interface, {}, module))
                continue
            commands.extend(self._set_config(interface, each, module))

        return commands

    def _state_deleted(self, want, have):
        """The command generator when state is deleted
        :rtype: A list
        :returns: the commands necessary to remove the current configuration
                  of the provided objects
        """
        commands = []

        if want:
            for interface in want:
                interface["name"] = normalize_interface(interface["name"])
                for each in have:
                    if each["name"] == interface["name"] or interface["name"] in each["name"]:
                        break
                else:
                    continue
                interface = dict(name=interface["name"])
                commands.extend(self._clear_config(interface, each))
        else:
            for each in have:
                want = dict()
                commands.extend(self._clear_config(want, each))

        return commands

    def _handle_deprecated(self, config):
        if config.get("q_vlan"):
            config["qvlan"] = config.get("q_vlan")
            del config["q_vlan"]
        return config

    def _set_config(self, want, have, module):
        # Set the interface config based on the want and have config
        commands = []
        interface = "interface " + want["name"]
        l2_protocol_bool = False
        # Get the diff b/w want and have
        want = self._handle_deprecated(want)
        have = self._handle_deprecated(have)
        diff = dict_diff(have, want)
        if diff:
            # For merging with already configured l2protocol
            if have.get("l2protocol") and len(have.get("l2protocol")) > 1:
                l2_protocol_diff = []
                for each in want.get("l2protocol"):
                    for every in have.get("l2protocol"):
                        if every == each:
                            break
                    if each not in have.get("l2protocol"):
                        l2_protocol_diff.append(each)
                l2_protocol_bool = True
                l2protocol = tuple(l2_protocol_diff)
            else:
                l2protocol = {}

            wants_native = diff.get("native_vlan")
            l2transport = diff.get("l2transport")
            qvlan = diff.get("qvlan")
            encapsulation = diff.get("encapsulation")
            propagate = diff.get("propagate")
            if l2_protocol_bool is False:
                l2protocol = diff.get("l2protocol")

            os_version = get_os_version(self._module)
            if os_version and Version(os_version) < Version("7.0.0"):
                if wants_native:
                    cmd = "dot1q native vlan {0}".format(wants_native)
                    add_command_to_config_list(interface, cmd, commands)

                if l2transport or l2protocol:
                    for each in l2protocol:
                        each = dict(each)
                        if isinstance(each, dict) and list(each.keys())[0] != "cpsv":
                            cmd = "l2transport l2protocol {0} {1}".format(
                                list(each.keys())[0],
                                list(each.values())[0],
                            )
                        add_command_to_config_list(interface, cmd, commands)

                if qvlan and "." in interface:
                    q_vlans = " ".join(map(str, want.get("qvlan")))
                    if q_vlans != have.get("qvlan"):
                        cmd = "dot1q vlan {0}".format(q_vlans)
                        add_command_to_config_list(interface, cmd, commands)
            else:
                if l2transport or l2protocol:
                    for each in l2protocol:
                        each = dict(each)
                        if isinstance(each, dict) and each.get("cpsv"):
                            cmd = "l2transport l2protocol {0} {1}".format(
                                "cpsv",
                                each.get("cpsv"),
                            )
                            add_command_to_config_list(
                                interface,
                                cmd,
                                commands,
                            )
                            break
                if encapsulation:
                    encapsulation = dict(encapsulation)
                    if encapsulation.get("dot1q"):
                        if encapsulation.get("second_dot1q"):
                            cmd = "encapsulation dot1q {0} second-dot1q {1}".format(
                                encapsulation.get("dot1q"),
                                encapsulation.get("second_dot1q"),
                            )
                        else:
                            cmd = "encapsulation dot1q {0}".format(
                                encapsulation.get("dot1q"),
                            )
                        add_command_to_config_list(interface, cmd, commands)

            if l2transport or l2protocol:
                if propagate and not have.get("propagate"):
                    cmd = "l2transport propagate remote-status"
                    add_command_to_config_list(interface, cmd, commands)
            elif want.get("l2transport") is False and (
                want.get("l2protocol") or want.get("propagate")
            ):
                module.fail_json(
                    msg="L2transport L2protocol or Propagate can only be configured when "
                    "L2transport set to True!",
                )

        return commands

    def _clear_config(self, want, have):
        # Delete the interface config based on the want and have config
        commands = []
        want = self._handle_deprecated(want)
        have = self._handle_deprecated(have)

        if want.get("name"):
            interface = "interface " + want["name"]
        else:
            interface = "interface " + have["name"]
        os_version = get_os_version(self._module)
        if os_version and Version(os_version) < Version("7.0.0"):
            if have.get("native_vlan"):
                remove_command_from_config_list(
                    interface,
                    "dot1q native vlan",
                    commands,
                )

            if have.get("qvlan"):
                remove_command_from_config_list(
                    interface,
                    "encapsulation dot1q",
                    commands,
                )
        else:
            if have.get("encapsulation"):
                remove_command_from_config_list(
                    interface,
                    "encapsulation dot1q",
                    commands,
                )

        if have.get("l2protocol") and (
            want.get("l2protocol") is None or want.get("propagate") is None
        ):
            if "no l2transport" not in commands:
                remove_command_from_config_list(
                    interface,
                    "l2transport",
                    commands,
                )
        elif have.get("l2transport") and have.get("l2transport") != want.get(
            "l2transport",
        ):
            if "no l2transport" not in commands:
                remove_command_from_config_list(
                    interface,
                    "l2transport",
                    commands,
                )
        return commands



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: lambda
version_added: 5.0.0
short_description: Manage AWS Lambda functions
description:
  - Allows for the management of Lambda functions.
  - This module was originally added to C(community.aws) in release 1.0.0.
options:
  name:
    description:
      - The name you want to assign to the function you are uploading. Cannot be changed.
    required: true
    type: str
  state:
    description:
      - Create or delete Lambda function.
    default: present
    choices: [ 'present', 'absent' ]
    type: str
  runtime:
    description:
      - The runtime environment for the Lambda function you are uploading.
      - Required when creating a function. Uses parameters as described in boto3 docs.
      - Required when I(state=present).
      - For supported list of runtimes, see U(https://docs.aws.amazon.com/lambda/latest/dg/lambda-runtimes.html).
    type: str
  role:
    description:
      - The Amazon Resource Name (ARN) of the IAM role that Lambda assumes when it executes your function to access any other Amazon Web Services (AWS)
        resources. You may use the bare ARN if the role belongs to the same AWS account.
      - Required when I(state=present).
    type: str
  handler:
    description:
      - The function within your code that Lambda calls to begin execution.
    type: str
  zip_file:
    description:
      - A .zip file containing your deployment package
      - If I(state=present) then either I(zip_file) or I(s3_bucket) must be present.
    aliases: [ 'src' ]
    type: str
  s3_bucket:
    description:
      - Amazon S3 bucket name where the .zip file containing your deployment package is stored.
      - If I(state=present) then either I(zip_file) or I(s3_bucket) must be present.
      - I(s3_bucket) and I(s3_key) are required together.
    type: str
  s3_key:
    description:
      - The Amazon S3 object (the deployment package) key name you want to upload.
      - I(s3_bucket) and I(s3_key) are required together.
    type: str
  s3_object_version:
    description:
      - The Amazon S3 object (the deployment package) version you want to upload.
    type: str
  description:
    description:
      - A short, user-defined function description. Lambda does not use this value. Assign a meaningful description as you see fit.
    type: str
    default: ''
  timeout:
    description:
      - The function maximum execution time in seconds after which Lambda should terminate the function.
    default: 3
    type: int
  memory_size:
    description:
      - The amount of memory, in MB, your Lambda function is given.
    default: 128
    type: int
  vpc_subnet_ids:
    description:
      - List of subnet IDs to run Lambda function in.
      - Use this option if you need to access resources in your VPC. Leave empty if you don't want to run the function in a VPC.
      - If set, I(vpc_security_group_ids) must also be set.
    type: list
    elements: str
  vpc_security_group_ids:
    description:
      - List of VPC security group IDs to associate with the Lambda function.
      - Required when I(vpc_subnet_ids) is used.
    type: list
    elements: str
  environment_variables:
    description:
      - A dictionary of environment variables the Lambda function is given.
    type: dict
  dead_letter_arn:
    description:
      - The parent object that contains the target Amazon Resource Name (ARN) of an Amazon SQS queue or Amazon SNS topic.
    type: str
  tracing_mode:
    description:
      - Set mode to 'Active' to sample and trace incoming requests with AWS X-Ray. Turned off (set to 'PassThrough') by default.
    choices: ['Active', 'PassThrough']
    type: str
  kms_key_arn:
    description:
      - The KMS key ARN used to encrypt the function's environment variables.
    type: str
    version_added: 3.3.0
    version_added_collection: community.aws
  architecture:
    description:
      - The instruction set architecture that the function supports.
      - Requires one of I(s3_bucket) or I(zip_file).
    type: str
    choices: ['x86_64', 'arm64']
    aliases: ['architectures']
    version_added: 5.0.0
  layers:
    description:
      - A list of function layers to add to the function's execution environment.
      - Specify each layer by its ARN, including the version.
    suboptions:
        layer_version_arn:
            description:
            - The ARN of the layer version.
            - Mutually exclusive with I(layer_version_arn).
            type: str
        layer_name:
            description:
            - The name or Amazon Resource Name (ARN) of the layer.
            - Mutually exclusive with I(layer_version_arn).
            type: str
            aliases: ['layer_arn']
        version:
            description:
            - The version number.
            - Required when I(layer_name) is provided, ignored if not.
            type: int
            aliases: ['layer_version']
    type: list
    elements: dict
    version_added: 5.5.0
  image_uri:
    description:
      - The Amazon ECR URI of the image to use.
      - Required (alternative to runtime zip_file and s3_bucket) when creating a function.
      - Required when I(state=present).
    type: str
    version_added: 7.3.0
author:
  - 'Steyn Huizinga (@steynovich)'
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""
# Create Lambda functions
- name: looped creation
  amazon.aws.lambda:
    name: '{{ item.name }}'
    state: present
    zip_file: '{{ item.zip_file }}'
    runtime: 'python2.7'
    role: 'arn:aws:iam::123456789012:role/lambda_basic_execution'
    handler: 'hello_python.my_handler'
    vpc_subnet_ids:
      - subnet-123abcde
      - subnet-edcba321
    vpc_security_group_ids:
      - sg-123abcde
      - sg-edcba321
    environment_variables: '{{ item.env_vars }}'
    tags:
      key1: 'value1'
  loop:
    - name: HelloWorld
      zip_file: hello-code.zip
      env_vars:
        key1: "first"
        key2: "second"
    - name: ByeBye
      zip_file: bye-code.zip
      env_vars:
        key1: "1"
        key2: "2"

# To remove previously added tags pass an empty dict
- name: remove tags
  amazon.aws.lambda:
    name: 'Lambda function'
    state: present
    zip_file: 'code.zip'
    runtime: 'python2.7'
    role: 'arn:aws:iam::123456789012:role/lambda_basic_execution'
    handler: 'hello_python.my_handler'
    tags: {}

# Basic Lambda function deletion
- name: Delete Lambda functions HelloWorld and ByeBye
  amazon.aws.lambda:
    name: '{{ item }}'
    state: absent
  loop:
    - HelloWorld
    - ByeBye

# Create Lambda functions with function layers
- name: looped creation
  amazon.aws.lambda:
    name: 'HelloWorld'
    state: present
    zip_file: 'hello-code.zip'
    runtime: 'python2.7'
    role: 'arn:aws:iam::123456789012:role/lambda_basic_execution'
    handler: 'hello_python.my_handler'
    layers:
      - layer_version_arn: 'arn:aws:lambda:us-east-1:123456789012:layer:python27-env:7'
"""

RETURN = r"""
code:
    description: The lambda function's code returned by get_function in boto3.
    returned: success
    type: dict
    contains:
        location:
            description:
                - The presigned URL you can use to download the function's .zip file that you previously uploaded.
                - The URL is valid for up to 10 minutes.
            returned: success
            type: str
            sample: 'https://prod-04-2014-tasks.s3.us-east-1.amazonaws.com/snapshots/sample'
        repository_type:
            description: The repository from which you can download the function.
            returned: success
            type: str
            sample: 'S3'
configuration:
    description: the lambda function's configuration metadata returned by get_function in boto3
    returned: success
    type: dict
    contains:
        architectures:
            description: The architectures supported by the function.
            type: list
            elements: str
            sample: ['arm64']
        code_sha256:
            description: The SHA256 hash of the function's deployment package.
            returned: success
            type: str
            sample: 'zOAGfF5JLFuzZoSNirUtOrQp+S341IOA3BcoXXoaIaU='
        code_size:
            description: The size of the function's deployment package in bytes.
            returned: success
            type: int
            sample: 123
        dead_letter_config:
            description: The function's dead letter queue.
            returned: when the function has a dead letter queue configured
            type: dict
            sample: { 'target_arn': arn:aws:lambda:us-east-1:123456789012:function:myFunction:1 }
            contains:
                target_arn:
                    description: The ARN of an SQS queue or SNS topic.
                    returned: when the function has a dead letter queue configured
                    type: str
                    sample: arn:aws:lambda:us-east-1:123456789012:function:myFunction:1
        description:
            description: The function's description.
            returned: success
            type: str
            sample: 'My function'
        environment:
            description: The function's environment variables.
            returned: when environment variables exist
            type: dict
            contains:
                variables:
                    description: Environment variable key-value pairs.
                    returned: when environment variables exist
                    type: dict
                    sample: {'key': 'value'}
                error:
                    description: Error message for environment variables that could not be applied.
                    returned: when there is an error applying environment variables
                    type: dict
                    contains:
                        error_code:
                            description: The error code.
                            returned: when there is an error applying environment variables
                            type: str
                        message:
                            description: The error message.
                            returned: when there is an error applying environment variables
                            type: str
        function_arn:
            description: The function's Amazon Resource Name (ARN).
            returned: on success
            type: str
            sample: 'arn:aws:lambda:us-east-1:123456789012:function:myFunction:1'
        function_name:
            description: The function's name.
            returned: on success
            type: str
            sample: 'myFunction'
        handler:
            description: The function Lambda calls to begin executing your function.
            returned: on success
            type: str
            sample: 'index.handler'
        last_modified:
            description: The date and time that the function was last updated, in ISO-8601 format (YYYY-MM-DDThh:mm:ssTZD).
            returned: on success
            type: str
            sample: '2017-08-01T00:00:00.000+0000'
        memory_size:
            description: The memory allocated to the function.
            returned: on success
            type: int
            sample: 128
        revision_id:
            description: The latest updated revision of the function or alias.
            returned: on success
            type: str
            sample: 'a2x9886d-d48a-4a0c-ab64-82abc005x80c'
        role:
            description: The function's execution role.
            returned: on success
            type: str
            sample: 'arn:aws:iam::123456789012:role/lambda_basic_execution'
        runtime:
            description: The funtime environment for the Lambda function.
            returned: on success
            type: str
            sample: 'nodejs6.10'
        tracing_config:
            description: The function's AWS X-Ray tracing configuration.
            returned: on success
            type: dict
            sample: { 'mode': 'Active' }
            contains:
                mode:
                    description: The tracing mode.
                    returned: on success
                    type: str
                    sample: 'Active'
        timeout:
            description: The amount of time that Lambda allows a function to run before terminating it.
            returned: on success
            type: int
            sample: 3
        version:
            description: The version of the Lambda function.
            returned: on success
            type: str
            sample: '1'
        vpc_config:
            description: The function's networking configuration.
            returned: on success
            type: dict
            sample: {
              'security_group_ids': [],
              'subnet_ids': [],
              'vpc_id': '123'
            }
        layers:
            description: The function's layers.
            returned: on success
            version_added: 5.5.0
            type: complex
            contains:
                arn:
                    description: The Amazon Resource Name (ARN) of the function layer.
                    returned: always
                    type: str
                    sample: active
                code_size:
                    description: The size of the layer archive in bytes.
                    returned: always
                    type: str
                signing_profile_version_arn:
                    description: The Amazon Resource Name (ARN) for a signing profile version.
                    returned: always
                    type: str
                signing_job_arn:
                    description: The Amazon Resource Name (ARN) of a signing job.
                    returned: always
                    type: str
"""

import base64
import hashlib
import re
import traceback
from collections import Counter

try:
    from botocore.exceptions import BotoCoreError
    from botocore.exceptions import ClientError
    from botocore.exceptions import WaiterError
except ImportError:
    pass  # protected by AnsibleAWSModule

from ansible.module_utils.common.dict_transformations import camel_dict_to_snake_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.iam import get_aws_account_info
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.tagging import compare_aws_tags


def get_current_function(connection, function_name, qualifier=None):
    try:
        if qualifier is not None:
            return connection.get_function(FunctionName=function_name, Qualifier=qualifier, aws_retry=True)
        return connection.get_function(FunctionName=function_name, aws_retry=True)
    except is_boto3_error_code("ResourceNotFoundException"):
        return None


def get_layer_version_arn(module, connection, layer_name, version_number):
    try:
        layer_versions = connection.list_layer_versions(LayerName=layer_name, aws_retry=True)["LayerVersions"]
        for v in layer_versions:
            if v["Version"] == version_number:
                return v["LayerVersionArn"]
        module.fail_json(msg=f"Unable to find version {version_number} from Lambda layer {layer_name}")
    except is_boto3_error_code("ResourceNotFoundException"):
        module.fail_json(msg=f"Lambda layer {layer_name} not found")


def sha256sum(filename):
    hasher = hashlib.sha256()
    with open(filename, "rb") as f:
        hasher.update(f.read())

    code_hash = hasher.digest()
    code_b64 = base64.b64encode(code_hash)
    hex_digest = code_b64.decode("utf-8")

    return hex_digest


def set_tag(client, module, tags, function, purge_tags):
    if tags is None:
        return False

    changed = False
    arn = function["Configuration"]["FunctionArn"]

    try:
        current_tags = client.list_tags(Resource=arn, aws_retry=True).get("Tags", {})
    except (BotoCoreError, ClientError) as e:
        module.fail_json_aws(e, msg="Unable to list tags")

    tags_to_add, tags_to_remove = compare_aws_tags(current_tags, tags, purge_tags=purge_tags)

    if not tags_to_remove and not tags_to_add:
        return False

    if module.check_mode:
        return True

    try:
        if tags_to_remove:
            client.untag_resource(
                Resource=arn,
                TagKeys=tags_to_remove,
                aws_retry=True,
            )
            changed = True

        if tags_to_add:
            client.tag_resource(
                Resource=arn,
                Tags=tags_to_add,
                aws_retry=True,
            )
            changed = True

    except (BotoCoreError, ClientError) as e:
        module.fail_json_aws(e, msg=f"Unable to tag resource {arn}")

    return changed


def wait_for_lambda(client, module, name):
    try:
        client_active_waiter = client.get_waiter("function_active")
        client_updated_waiter = client.get_waiter("function_updated")
        client_active_waiter.wait(FunctionName=name)
        client_updated_waiter.wait(FunctionName=name)
    except WaiterError as e:
        module.fail_json_aws(e, msg="Timeout while waiting on lambda to finish updating")
    except (ClientError, BotoCoreError) as e:
        module.fail_json_aws(e, msg="Failed while waiting on lambda to finish updating")


def format_response(response):
    tags = response.get("Tags", {})
    result = camel_dict_to_snake_dict(response)
    # Lambda returns a dict rather than the usual boto3 list of dicts
    result["tags"] = tags
    return result


def _zip_args(zip_file, current_config, ignore_checksum):
    if not zip_file:
        return {}

    # If there's another change that needs to happen, we always re-upload the code
    if not ignore_checksum:
        local_checksum = sha256sum(zip_file)
        remote_checksum = current_config.get("CodeSha256", "")
        if local_checksum == remote_checksum:
            return {}

    with open(zip_file, "rb") as f:
        zip_content = f.read()
    return {"ZipFile": zip_content}


def _s3_args(s3_bucket, s3_key, s3_object_version):
    if not s3_bucket:
        return {}
    if not s3_key:
        return {}

    code = {"S3Bucket": s3_bucket, "S3Key": s3_key}
    if s3_object_version:
        code.update({"S3ObjectVersion": s3_object_version})

    return code


def _image_args(image_uri):
    if not image_uri:
        return {}

    code = {"ImageUri": image_uri}
    return code


def _code_args(module, current_config):
    s3_bucket = module.params.get("s3_bucket")
    s3_key = module.params.get("s3_key")
    s3_object_version = module.params.get("s3_object_version")
    zip_file = module.params.get("zip_file")
    architectures = module.params.get("architecture")
    image_uri = module.params.get("image_uri")

    code_kwargs = {}

    if architectures and current_config.get("Architectures", None) != [architectures]:
        module.warn("Arch Change")
        code_kwargs.update({"Architectures": [architectures]})

    try:
        code_kwargs.update(_zip_args(zip_file, current_config, bool(code_kwargs)))
    except IOError as e:
        module.fail_json(msg=str(e), exception=traceback.format_exc())

    code_kwargs.update(_s3_args(s3_bucket, s3_key, s3_object_version))
    code_kwargs.update(_image_args(image_uri))

    if not code_kwargs:
        return {}

    if not architectures and current_config.get("Architectures", None):
        code_kwargs.update({"Architectures": current_config.get("Architectures", None)})

    return code_kwargs


def main():
    argument_spec = dict(
        name=dict(required=True),
        state=dict(default="present", choices=["present", "absent"]),
        image_uri=dict(),
        runtime=dict(),
        role=dict(),
        handler=dict(),
        zip_file=dict(aliases=["src"]),
        s3_bucket=dict(),
        s3_key=dict(no_log=False),
        s3_object_version=dict(),
        description=dict(default=""),
        timeout=dict(type="int", default=3),
        memory_size=dict(type="int", default=128),
        vpc_subnet_ids=dict(type="list", elements="str"),
        vpc_security_group_ids=dict(type="list", elements="str"),
        environment_variables=dict(type="dict"),
        dead_letter_arn=dict(),
        kms_key_arn=dict(type="str", no_log=False),
        tracing_mode=dict(choices=["Active", "PassThrough"]),
        architecture=dict(choices=["x86_64", "arm64"], type="str", aliases=["architectures"]),
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(type="bool", default=True),
        layers=dict(
            type="list",
            elements="dict",
            options=dict(
                layer_version_arn=dict(type="str"),
                layer_name=dict(type="str", aliases=["layer_arn"]),
                version=dict(type="int", aliases=["layer_version"]),
            ),
            required_together=[["layer_name", "version"]],
            required_one_of=[["layer_version_arn", "layer_name"]],
            mutually_exclusive=[["layer_name", "layer_version_arn"], ["version", "layer_version_arn"]],
        ),
    )

    mutually_exclusive = [
        ["zip_file", "s3_key"],
        ["zip_file", "s3_bucket"],
        ["zip_file", "s3_object_version"],
        ["image_uri", "zip_file"],
        ["image_uri", "runtime"],
        ["image_uri", "handler"],
        ["image_uri", "s3_key"],
        ["image_uri", "s3_bucket"],
        ["image_uri", "s3_object_version"],
    ]

    required_by = {"runtime": ["handler"]}

    required_together = [
        ["s3_key", "s3_bucket"],
        ["vpc_subnet_ids", "vpc_security_group_ids"],
        ["runtime", "handler"],
    ]

    required_if = [
        ["state", "present", ["role"]],
        ["state", "present", ["runtime", "image_uri"], True],
        ["architecture", "x86_64", ["zip_file", "s3_bucket", "image_uri"], True],
        ["architecture", "arm64", ["zip_file", "s3_bucket", "image_uri"], True],
    ]

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=mutually_exclusive,
        required_together=required_together,
        required_if=required_if,
    )

    name = module.params.get("name")
    state = module.params.get("state").lower()
    runtime = module.params.get("runtime")
    role = module.params.get("role")
    handler = module.params.get("handler")
    description = module.params.get("description")
    timeout = module.params.get("timeout")
    memory_size = module.params.get("memory_size")
    vpc_subnet_ids = module.params.get("vpc_subnet_ids")
    vpc_security_group_ids = module.params.get("vpc_security_group_ids")
    environment_variables = module.params.get("environment_variables")
    dead_letter_arn = module.params.get("dead_letter_arn")
    tracing_mode = module.params.get("tracing_mode")
    tags = module.params.get("tags")
    purge_tags = module.params.get("purge_tags")
    kms_key_arn = module.params.get("kms_key_arn")
    architectures = module.params.get("architecture")
    image_uri = module.params.get("image_uri")
    layers = []

    check_mode = module.check_mode
    changed = False

    try:
        client = module.client("lambda", retry_decorator=AWSRetry.jittered_backoff())
    except (ClientError, BotoCoreError) as e:
        module.fail_json_aws(e, msg="Trying to connect to AWS")

    if state == "present":
        if re.match(r"^arn:aws(-([a-z\-]+))?:iam", role):
            role_arn = role
        else:
            # get account ID and assemble ARN
            account_id, partition = get_aws_account_info(module)
            role_arn = f"arn:{partition}:iam::{account_id}:role/{role}"

        # create list of layer version arn
        if module.params.get("layers"):
            for layer in module.params.get("layers"):
                layer_version_arn = layer.get("layer_version_arn")
                if layer_version_arn is None:
                    layer_version_arn = get_layer_version_arn(
                        module, client, layer.get("layer_name"), layer.get("version")
                    )
                layers.append(layer_version_arn)

    # Get function configuration if present, False otherwise
    current_function = get_current_function(client, name)

    # Update existing Lambda function
    if state == "present" and current_function:
        # Get current state
        current_config = current_function["Configuration"]
        current_version = None

        # Update function configuration
        func_kwargs = {"FunctionName": name}

        # Update configuration if needed
        if role_arn and current_config["Role"] != role_arn:
            func_kwargs.update({"Role": role_arn})
        if handler and current_config["Handler"] != handler:
            func_kwargs.update({"Handler": handler})
        if description and current_config["Description"] != description:
            func_kwargs.update({"Description": description})
        if timeout and current_config["Timeout"] != timeout:
            func_kwargs.update({"Timeout": timeout})
        if memory_size and current_config["MemorySize"] != memory_size:
            func_kwargs.update({"MemorySize": memory_size})
        if image_uri is not None and current_config["PackageType"] != "Image":
            func_kwargs.update({"PackageType": "Image"})
        if runtime and current_config["Runtime"] != runtime:
            func_kwargs.update({"Runtime": runtime})
        if (environment_variables is not None) and (
            current_config.get("Environment", {}).get("Variables", {}) != environment_variables
        ):
            func_kwargs.update({"Environment": {"Variables": environment_variables}})
        if dead_letter_arn is not None:
            if current_config.get("DeadLetterConfig"):
                if current_config["DeadLetterConfig"]["TargetArn"] != dead_letter_arn:
                    func_kwargs.update({"DeadLetterConfig": {"TargetArn": dead_letter_arn}})
            else:
                if dead_letter_arn != "":
                    func_kwargs.update({"DeadLetterConfig": {"TargetArn": dead_letter_arn}})
        if tracing_mode and (current_config.get("TracingConfig", {}).get("Mode", "PassThrough") != tracing_mode):
            func_kwargs.update({"TracingConfig": {"Mode": tracing_mode}})
        if kms_key_arn:
            func_kwargs.update({"KMSKeyArn": kms_key_arn})

        # If VPC configuration is desired
        if vpc_subnet_ids:
            if "VpcConfig" in current_config:
                # Compare VPC config with current config
                current_vpc_subnet_ids = current_config["VpcConfig"]["SubnetIds"]
                current_vpc_security_group_ids = current_config["VpcConfig"]["SecurityGroupIds"]

                subnet_net_id_changed = sorted(vpc_subnet_ids) != sorted(current_vpc_subnet_ids)
                vpc_security_group_ids_changed = sorted(vpc_security_group_ids) != sorted(
                    current_vpc_security_group_ids
                )

            if "VpcConfig" not in current_config or subnet_net_id_changed or vpc_security_group_ids_changed:
                new_vpc_config = {"SubnetIds": vpc_subnet_ids, "SecurityGroupIds": vpc_security_group_ids}
                func_kwargs.update({"VpcConfig": new_vpc_config})
        else:
            # No VPC configuration is desired, assure VPC config is empty when present in current config
            if "VpcConfig" in current_config and current_config["VpcConfig"].get("VpcId"):
                func_kwargs.update({"VpcConfig": {"SubnetIds": [], "SecurityGroupIds": []}})

        # Check layers
        if layers:
            # compare two lists to see if the target layers are equal to the current
            current_layers = current_config.get("Layers", [])
            if Counter(layers) != Counter((f["Arn"] for f in current_layers)):
                func_kwargs.update({"Layers": layers})

        # Upload new configuration if configuration has changed
        if len(func_kwargs) > 1:
            if not check_mode:
                wait_for_lambda(client, module, name)

            try:
                if not check_mode:
                    response = client.update_function_configuration(aws_retry=True, **func_kwargs)
                    current_version = response["Version"]
                changed = True
            except (BotoCoreError, ClientError) as e:
                module.fail_json_aws(e, msg="Trying to update lambda configuration")

        # Tag Function
        if tags is not None:
            if set_tag(client, module, tags, current_function, purge_tags):
                changed = True

        code_kwargs = _code_args(module, current_config)
        if code_kwargs:
            # Update code configuration
            code_kwargs.update({"FunctionName": name, "Publish": True})

            if not check_mode:
                wait_for_lambda(client, module, name)

            try:
                if not check_mode:
                    response = client.update_function_code(aws_retry=True, **code_kwargs)
                    current_version = response["Version"]
                changed = True
            except (BotoCoreError, ClientError) as e:
                module.fail_json_aws(e, msg="Trying to upload new code")

        # Describe function code and configuration
        response = get_current_function(client, name, qualifier=current_version)
        if not response:
            module.fail_json(msg="Unable to get function information after updating")
        response = format_response(response)
        # We're done
        module.exit_json(changed=changed, code_kwargs=code_kwargs, func_kwargs=func_kwargs, **response)

    # Function doesn't exist, create new Lambda function
    elif state == "present":
        func_kwargs = {
            "FunctionName": name,
            "Publish": True,
            "Role": role_arn,
            "Timeout": timeout,
            "MemorySize": memory_size,
        }

        code = _code_args(module, {})
        if not code:
            module.fail_json(msg="Either S3 object or path to zipfile required")
        if "Architectures" in code:
            func_kwargs.update({"Architectures": code.pop("Architectures")})
        func_kwargs.update({"Code": code})

        if description is not None:
            func_kwargs.update({"Description": description})

        if image_uri is not None:
            func_kwargs.update({"PackageType": "Image"})

        if runtime is not None:
            func_kwargs.update({"Runtime": runtime})

        if handler is not None:
            func_kwargs.update({"Handler": handler})

        if environment_variables:
            func_kwargs.update({"Environment": {"Variables": environment_variables}})

        if dead_letter_arn:
            func_kwargs.update({"DeadLetterConfig": {"TargetArn": dead_letter_arn}})

        if tracing_mode:
            func_kwargs.update({"TracingConfig": {"Mode": tracing_mode}})

        if kms_key_arn:
            func_kwargs.update({"KMSKeyArn": kms_key_arn})

        # If VPC configuration is given
        if vpc_subnet_ids:
            func_kwargs.update({"VpcConfig": {"SubnetIds": vpc_subnet_ids, "SecurityGroupIds": vpc_security_group_ids}})

        # Layers
        if layers:
            func_kwargs.update({"Layers": layers})

        # Tag Function
        if tags:
            func_kwargs.update({"Tags": tags})

        # Function would have been created if not check mode
        if check_mode:
            module.exit_json(changed=True)

        # Finally try to create function
        current_version = None
        try:
            response = client.create_function(aws_retry=True, **func_kwargs)
            current_version = response["Version"]
            changed = True
        except (BotoCoreError, ClientError) as e:
            module.fail_json_aws(e, msg="Trying to create function")

        response = get_current_function(client, name, qualifier=current_version)
        if not response:
            module.fail_json(msg="Unable to get function information after creating")
        response = format_response(response)
        module.exit_json(changed=changed, **response)

    # Delete existing Lambda function
    if state == "absent" and current_function:
        try:
            if not check_mode:
                client.delete_function(FunctionName=name, aws_retry=True)
            changed = True
        except (BotoCoreError, ClientError) as e:
            module.fail_json_aws(e, msg="Trying to delete Lambda function")

        module.exit_json(changed=changed)

    # Function already absent, do nothing
    elif state == "absent":
        module.exit_json(changed=changed)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The L3_interfaces parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def ip_tmplt(config_data):
    cmd = "ipv6 address {ip}"
    if config_data.get("ipv6"):
        config = config_data.get("ipv6")
        cmd = cmd.format(ip=config["address"])
    if config.get("segment_routing"):
        # if config.get("segment_routing").get("enable"):
        cmd += " segment-routing"
        if config.get("segment_routing").get("default"):
            cmd += " default"
        if config.get("segment_routing").get("ipv6_sr"):
            cmd += " ipv6-sr"
    if config.get("secondary"):
        cmd += " secondary"
    if config.get("link_local"):
        cmd += " link-local"
    if config.get("anycast"):
        cmd += " anycast"
    if config.get("cga"):
        cmd += " cga"
    if config.get("eui"):
        cmd += " eui"
    return cmd


def ipv4_dhcp(config_data):
    _data = config_data.get("ipv4", {}).get("dhcp")
    if not _data.get("enable", True):
        return ""
    cmd = "ip address dhcp"
    if _data.get("client_id"):
        cmd += " client-id {client_id}".format(**_data)
    if _data.get("hostname"):
        cmd += " hostname {hostname}".format(**_data)
    return cmd


class L3_interfacesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(L3_interfacesTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "name",
            "getval": re.compile(
                r"""^interface
                    (\s(?P<name>\S+))
                    $""",
                re.VERBOSE,
            ),
            "compval": "name",
            "setval": "interface {{ name }}",
            "result": {"{{ name }}": {"name": "{{ name }}"}},
            "shared": True,
        },
        {
            "name": "ipv4.address",
            "getval": re.compile(
                r"""\s+ip\saddress
                    (\s(?P<ipv4>\S+))
                    (\s(?P<netmask>\S+))
                    (\s(?P<secondary>secondary))?
                    $""",
                re.VERBOSE,
            ),
            "setval": "ip address {{ ipv4.address }}"
            "{{ ' secondary' if ipv4.secondary|d(False) else ''}}",
            "result": {
                "{{ name }}": {
                    "ipv4": [
                        {
                            "address": "{{ ipv4 }}",
                            "netmask": "{{ netmask }}",
                            "secondary": "{{ True if secondary is defined }}",
                        },
                    ],
                },
            },
        },
        {
            "name": "ipv4.pool",
            "getval": re.compile(
                r"""
                \s+ip\saddress\spool\s(?P<pool>.+$)
                $""", re.VERBOSE,
            ),
            "setval": "ip address pool {{ ipv4.pool }}",
            "result": {
                "{{ name }}": {
                    "ipv4": [
                        {
                            "pool": "{{ pool }}",
                        },
                    ],
                },
            },
        },
        {
            "name": "ipv4.dhcp",
            "getval": re.compile(
                r"""\s+ip\saddress\s
                    ((?P<dhcp>dhcp))
                    (\sclient-id\s(?P<client_id>\S+))?
                    (\shostname\s(?P<hostname>\S+))?
                    $""",
                re.VERBOSE,
            ),
            "setval": ipv4_dhcp,
            "result": {
                "{{ name }}": {
                    "ipv4": [
                        {
                            "dhcp": {
                                "enable": "{{ True if dhcp is defined }}",
                                "client_id": "{{ client_id }}",
                                "hostname": "{{ hostname }}",
                            },
                        },
                    ],
                },
            },
        },
        {
            "name": "ipv6.address",
            "getval": re.compile(
                r"""\s+ipv6\saddress
                    (\s(?P<ipv6>\b(?!autoconfig\b)\S+))
                    (\s(?P<link_local>link-local))?
                    (\s(?P<anycast>anycast))?
                    (\s(?P<cga>cga))?
                    (\s(?P<eui>eui))?
                    (\s(?P<enable>segment-routing))?
                    (\s(?P<default>default))?
                    (\s(?P<ipv6_sr>ipv6-sr))?
                    $""",
                    re.VERBOSE,
            ),
            "setval": ip_tmplt,
            "result": {
                "{{ name }}": {
                    "ipv6": [
                        {
                            "address": "{{ ipv6 }}",
                            "link_local": "{{ True if link_local is defined }}",
                            "anycast": "{{ True if anycast is defined }}",
                            "cga": "{{ True if cga is defined }}",
                            "eui": "{{ True if eui is defined }}",
                            "segment_routing": {
                                "enable": "{{ True if enable is defined }}",
                                "default": "{{ True if default is defined }}",
                                "ipv6_sr": "{{ True if ipv6_sr is defined }}",
                            },
                        },
                    ],
                },
            },
        },
        {
            "name": "ipv6.autoconfig",
            "getval": re.compile(
                r"""\s+ipv6\saddress\s
                    ((?P<enable>autoconfig))
                    (\s(?P<default>default))?
                    $""",
                    re.VERBOSE,
            ),
            "setval": "{{ 'ipv6 address autoconfig' if ipv6.autoconfig.enable|d(False) or ipv6.autoconfig.default|d(False) else ''}}"
            "{{ ' default' if ipv6.autoconfig.default|d(False) else ''}}",
            "result": {
                "{{ name }}": {
                    "ipv6": [
                        {
                            "autoconfig": {
                                "enable": "{{ True if enable is defined }}",
                                "default": "{{ True if default is defined }}",
                            },
                        },
                    ],
                },
            },
        },
        {
            "name": "ipv6.dhcp",
            "getval": re.compile(
                r"""\s+ipv6\saddress\sdhcp
                    (\s(?P<rapid_commit>rapid-commit))?
                    $""",
                    re.VERBOSE,
            ),
            "setval": "{{ 'ipv6 address dhcp' if ipv6.dhcp.enable|d(False)|d(False) or ipv6.dhcp.rapid_commit|d(False)|d(False)else ''}}"
            "{{ ' rapid-commit' if ipv6.dhcp.rapid_commit|d(False) else ''}}",
            "result": {
                "{{ name }}": {
                    "ipv6": [
                        {
                            "dhcp": {
                                "enable": True,
                                "rapid_commit": "{{ True if rapid_commit is defined }}",
                            },
                        },
                    ],
                },
            },
        },
    ]
    # fmt: on



# Copyright (c) 2022, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import base64

from ansible import constants as C
from ansible.plugins.action import ActionBase
from ansible.utils.vars import merge_hash

from ansible_collections.community.docker.plugins.module_utils._scramble import unscramble


class ActionModule(ActionBase):
    # Set to True when transferring files to the remote
    TRANSFERS_FILES = False

    def run(self, tmp=None, task_vars=None):
        self._supports_check_mode = True
        self._supports_async = True

        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        self._task.args['_max_file_size_for_diff'] = C.MAX_FILE_SIZE_FOR_DIFF

        result = merge_hash(result, self._execute_module(task_vars=task_vars, wrap_async=self._task.async_val))

        if u'diff' in result and result[u'diff'].get(u'scrambled_diff'):
            # Scrambling is not done for security, but to avoid no_log screwing up the diff
            diff = result[u'diff']
            key = base64.b64decode(diff.pop(u'scrambled_diff'))
            for k in (u'before', u'after'):
                if k in diff:
                    diff[k] = unscramble(diff[k], key)

        return result
